from __future__ import annotations
from corkus.objects.playtime import PlayerPlaytime
from typing import List, Union
import iso8601
import datetime

from .base_player import BasePlayer
from .partial_member import PartialMember
from .partial_guild import PartialGuild
from .member import GuildRank
from .player_status import PlayerStatus
from .player_statistics import PlayerStatistics
from .player_class import PlayerClass
from .dungeon import Dungeon
from .raid import Raid
from .quest import Quest
from .player_ranking import Ranking

class Player(BasePlayer):
    """Represents a player of a Wynncraft server."""
    @property
    def username(self) -> str:
        """Minecraft username of player."""
        return self._attributes.get("username", "")

    @property
    def join_date(self) -> datetime.datetime:
        """Date and time when player joined Wynncraft first time."""
        return iso8601.parse_date(self._attributes.get("meta", {}).get("firstJoin", "1970"))

    @property
    def last_online(self) -> datetime.datetime:
        """Date and time when player was last seen online."""
        if self.status.online:
            return datetime.datetime.now(datetime.timezone.utc)
        else:
            return iso8601.parse_date(self._attributes.get("meta", {}).get("lastJoin", "1970"))

    @property
    def online(self) -> bool:
        """Is player online right now, for detailed information see :py:attr:`~status`."""
        return self.status.online

    @property
    def status(self) -> PlayerStatus:
        """Information about player's current online status."""
        return PlayerStatus(self._corkus, self._attributes.get("meta", {}).get("location", {}))

    @property
    def playtime(self) -> PlayerPlaytime:
        """Time that player spent on wynncraft servers."""
        return PlayerPlaytime(self._attributes.get("meta", {}).get("playtime", 0))

    @property
    def combined_level(self) -> int:
        """Combined level of all professions across all classes including combat."""
        return sum(c.combined_level for c in self.classes)

    @property
    def classes(self) -> List[PlayerClass]:
        """All of the player's classes."""
        return [PlayerClass(self._corkus, self, c) for c in self._attributes.get("classes", {})]

    @property
    def member(self) -> Union[PartialMember, None]:
        """Partial representation of player in the guild."""
        if self._attributes.get("guild", {}).get("name", None) is None:
            return None
        else:
            return PartialMember(
                corkus = self._corkus,
                uuid = self.uuid,
                username = self.username,
                guild = self.guild,
                rank = GuildRank(self._attributes.get("guild", {}).get("rank", "RECRUIT"))
            )

    @property
    def guild(self) -> Union[PartialGuild, None]:
        """Partial information about player's guild."""
        if self._attributes.get("guild", {}).get("name", None) is None:
            return None
        else:
            return PartialGuild(self._corkus, self._attributes.get("guild", {}).get("name", ""))

    @property
    def statistics(self) -> PlayerStatistics:
        """General statistics across all classes."""
        return PlayerStatistics(self._corkus, self._attributes.get("global", {}))

    @property
    def quests(self) -> List[Quest]:
        """List of all quests completed by player, combined across all classes."""
        result = []
        for quest in [q for c in self.classes for q in c.quests]:
            if any(q.name == quest for q in result): continue
            result.append(quest)
        return result

    @property
    def dungeons(self) -> List[Dungeon]:
        """List of dungeons completed by this player, combined across all classes."""
        result = []
        for dungeon in [d for c in self.classes for d in c.dungeons]:
            if any(q.name == dungeon for q in result): continue
            result.append(dungeon)
        return result

    @property
    def raids(self) -> List[Raid]:
        """List of raids completed by this player, combined across all classes."""
        result = []
        for raid in [r for c in self.classes for r in c.raids]:
            if any(q.name == raid for q in result): continue
            result.append(raid)
        return result

    @property
    def ranking(self) -> Ranking:
        """Place of player in various leaderboards.

        .. caution::
            This property will return invalid information if player name is
            improperly capitalized.
        """
        return Ranking(self._corkus, self._attributes.get("ranking", {}))

    def __repr__(self) -> str:
        return f"<Player username={self.username!r}>"



# type: ignore

from .network import NetworkEndpoint
from .player import PlayerEndpoint
from .guild import GuildEndpoint
from .territory import TerritoryEndpoint
from .search import SearchEndpoint
from .ingredient import IngredientEndpoint
from .leaderboard import LeaderboardEndpoint
from .recipe import RecipeEndpoint
from .item import ItemEndpoint



from __future__ import annotations
from typing import List, Literal
from .base import CorkusBase
from enum import Enum

class BannerColor(Enum):
    """List of colors available for :py:class:`GuildBanner` and :py:class:`GuildBannerLayer`."""
    WHITE = "WHITE"
    ORANGE = "ORANGE"
    MAGENTA = "MAGENTA"
    LIGHT_BLUE = "LIGHT_BLUE"
    YELLOW = "YELLOW"
    LIME = "LIME"
    PINK = "PINK"
    GRAY = "GRAY"
    LIGHT_GRAY = "LIGHT_GRAY"
    CYAN = "CYAN"
    PURPLE = "PURPLE"
    BLUE = "BLUE"
    BROWN = "BROWN"
    GREEN = "GREEN"
    RED = "RED"
    BLACK = "BLACK"

class BannerPattern(Enum):
    """List of patterns available for :py:class:`GuildBannerLayer`."""
    CIRCLE_MIDDLE = "CIRCLE_MIDDLE"
    SQUARE_BOTTOM_LEFT = "SQUARE_BOTTOM_LEFT"
    SQUARE_BOTTOM_RIGHT = "SQUARE_BOTTOM_RIGHT"
    SQUARE_TOP_LEFT = "SQUARE_TOP_LEFT"
    SQUARE_TOP_RIGHT = "SQUARE_TOP_RIGHT"
    HALF_HORIZONTAL = "HALF_HORIZONTAL"
    STRIPE_BOTTOM = "STRIPE_BOTTOM"
    STRIPE_TOP = "STRIPE_TOP"
    HALF_VERTICAL = "HALF_VERTICAL"
    STRIPE_LEFT = "STRIPE_LEFT"
    STRIPE_CENTER = "STRIPE_CENTER"
    STRIPE_RIGHT = "STRIPE_RIGHT"
    STRIPE_MIDDLE = "STRIPE_MIDDLE"
    STRAIGHT_CROSS = "STRAIGHT_CROSS"
    STRIPE_DOWNLEFT = "STRIPE_DOWNLEFT"
    STRIPE_DOWNRIGHT = "STRIPE_DOWNRIGHT"
    CROSS = "CROSS"
    DIAGONAL_LEFT = "DIAGONAL_LEFT"
    DIAGONAL_UP_RIGHT = "DIAGONAL_UP_RIGHT"
    TRIANGLE_TOP = "TRIANGLE_TOP"
    TRIANGLE_BOTTOM = "TRIANGLE_BOTTOM"
    RHOMBUS_MIDDLE = "RHOMBUS_MIDDLE"
    TRIANGLES_TOP = "TRIANGLES_TOP"
    TRIANGLES_BOTTOM = "TRIANGLES_BOTTOM"
    CURLY_BORDER = "CURLY_BORDER"
    BORDER = "BORDER"
    STRIPE_SMALL = "STRIPE_SMALL"
    BRICKS = "BRICKS"
    GRADIENT = "GRADIENT"
    CREEPER = "CREEPER"
    SKULL = "SKULL"
    FLOWER = "FLOWER"
    MOJANG = "MOJANG"
    DIAGONAL_LEFT_MIRROR = "DIAGONAL_LEFT_MIRROR"
    DIAGONAL_RIGHT = "DIAGONAL_RIGHT"
    GRADIENT_UP = "GRADIENT_UP"
    HALF_HORIZONTAL_MIRROR = "HALF_HORIZONTAL_MIRROR"
    HALF_VERTICAL_MIRROR = "HALF_VERTICAL_MIRROR"

class GuildBannerLayer(CorkusBase):
    """Represents a layer of :py:class:`GuildBanner`"""
    @property
    def color(self) -> BannerColor:
        """Color of the pattern."""
        return BannerColor(self._attributes.get("colour", "WHITE"))

    @property
    def pattern(self) -> BannerPattern:
        """Pattern used on that layer."""
        return BannerPattern(self._attributes.get("pattern", "CROSS"))

    def __repr__(self) -> str:
        return f"<GuildBannerLayer color={self.color.value!r} pattern={self.pattern.value!r}>"

class GuildBanner(CorkusBase):
    """:py:class:`Guild` banners are used to showcase a guild by means of a banner whenever
    they control a :py:class:`Territory`. Each banner is exclusive and thus no two guilds can have the same banner."""
    @property
    def base_color(self) -> BannerColor:
        """Color of the banner background."""
        return BannerColor(self._attributes.get("base", "WHITE"))

    @property
    def tier(self) -> Literal[1, 2, 3, 4, 5, 6]:
        """Tier of the banner represented by pedestal."""
        return self._attributes.get("tier", 1)

    @property
    def layers(self) -> List[GuildBannerLayer]:
        """List of banner layers."""
        return [GuildBannerLayer(self._corkus, l) for l in self._attributes.get("layers", {})]

    def __repr__(self) -> str:
        return f"<GuildBanner tier={self.tier} base_color={self.base_color.value!r} layers={self.layers}>"



from typing import Tuple

class Color:
    """Represents a color returned by API. This class is similar
    to a (red, green, blue) :py:class:`tuple`.
    """
    def __init__(self, rgb: Tuple[int, int, int]) -> None:
        self._value = rgb

    @property
    def r(self) -> int:
        """Return red component of the color."""
        return self._value[0]

    @property
    def g(self) -> int:
        """Return green component of the color."""
        return self._value[1]

    @property
    def b(self) -> int:
        """Return blue component of the color."""
        return self._value[2]

    @property
    def rgb(self) -> Tuple[int, int, int]:
        """Return (red, green, blue) :py:class:`tuple`"""
        return self._value

    @property
    def hex(self) -> str:
        """Return hex representation of color like ``#4e80f7``."""
        return '#%02x%02x%02x' % self._value

    def __str__(self) -> str:
        return self.hex



from __future__ import annotations
from typing import List, Optional

from .endpoint import Endpoint
from corkus.utils.request import APIVersion
from corkus.objects import Territory

class TerritoryEndpoint(Endpoint):
    async def list_all(self, timeout: Optional[int] = None) -> List[Territory]:
        """Get list of all possible teritories.

        :param timeout: Optionally override default timeout.
        """
        response = await self._request.get(
            version = APIVersion.V1,
            parameters = "territoryList",
            timeout = timeout
        )
        return [Territory(self._corkus, t) for t in response.get("territories", {}).values()]



from __future__ import annotations

from .base import CorkusBase

class ArmourDefence(CorkusBase):
    """Defence properties of :py:class:`Item`."""
    @property
    def fire(self) -> int:
        """Defence against fire element attacks."""
        return self._attributes.get("fireDefense", 0)

    @property
    def water(self) -> int:
        """Defence against water element attacks."""
        return self._attributes.get("waterDefense", 0)

    @property
    def air(self) -> int:
        """Defence against air element attacks."""
        return self._attributes.get("airDefense", 0)

    @property
    def thunder(self) -> int:
        """Defence against thunder element attacks."""
        return self._attributes.get("thunderDefense", 0)

    @property
    def earth(self) -> int:
        """Defence against earth element attacks."""
        return self._attributes.get("earthDefense", 0)

    def __repr__(self) -> str:
        return (
            f"<ArmourDefence fire={self.fire} water={self.water} "
            f"air={self.air} thunder={self.thunder} earth={self.earth}>"
        )



from __future__ import annotations
from typing import TYPE_CHECKING, Optional

from .base import CorkusBase
from .enums import ItemType
from .level_range import LevelRange

if TYPE_CHECKING:
    from .recipe import Recipe

class PartialRecipe(CorkusBase):
    """Represents a ``Partial`` version of :py:class:`Recipe`."""
    @property
    def id(self) -> str:
        """Return the id of recipe like ``Food-13-15`` or ``Wand-63-65``."""
        return self._attributes

    @property
    def type(self) -> ItemType:
        """Type of the item."""
        return ItemType(self.id.split("-")[0].upper())

    @property
    def level(self) -> LevelRange:
        """Level range that this item should be used in."""
        return LevelRange(self._corkus,
            min = int(self.id.split("-")[1].upper()),
            max = int(self.id.split("-")[2].upper())
        )

    async def fetch(self, timeout: Optional[int] = None) -> Recipe:
        """Fetch full recipe information from API.

        .. include:: ../note_api_call.rst

        :param timeout: Optionally override default timeout.
        """
        return await self._corkus.recipe.get_by_id(self.id, timeout)

    def __repr__(self) -> str:
        return f"<PartialRecipe id={self.id!r}>"



from __future__ import annotations
from corkus.objects.identification_type import IdentificationType
from typing import List, Literal, Optional, Tuple, Union

from ..utils.utils import Utils
from .endpoint import Endpoint
from corkus.utils.request import APIVersion
from corkus.objects import Ingredient, PartialIngredient, LogicSymbol, ProfessionType

class IngredientEndpoint(Endpoint):
    async def list_all(self, timeout: Optional[int] = None) -> List[PartialIngredient]:
        """List all available ingredient.

        :param timeout: Optionally override default timeout.
        """
        response = await self._request.get(
            version = APIVersion.V2,
            parameters = "ingredient/list",
            timeout = timeout
        )
        return [PartialIngredient(self._corkus, i) for i in response.get("data", [])]

    async def get(self, name: str, timeout: Optional[int] = None) -> Ingredient:
        """Get information about ingredient of specified name

        :param name: the name of the ingredient
        :param timeout: Optionally override default timeout.
        """
        response = await self._request.get(
            version = APIVersion.V2,
            parameters = "ingredient/get/" + name.replace(" ", "_"),
            timeout = timeout
        )
        return Ingredient(self._corkus, response.get("data", {})[0])

    async def search_by_name(self, name: str, timeout: Optional[int] = None) -> List[Ingredient]:
        """Search for the ingredients using name
        (:py:attr:`Ingredient.name <corkus.objects.Ingredient.name>`).
        For geting ingredient by specified name see: :py:func:`get`.

        :param name: Search term.
        :param timeout: Optionally override default timeout.
        """
        response = await self._request.get(
            version = APIVersion.V2,
            parameters = "ingredient/search/name/" + name,
            timeout = timeout
        )
        return [Ingredient(self._corkus, i) for i in response.get("data", [])]

    async def search_by_tier(self, tier: Literal[0, 1, 2, 3], timeout: Optional[int] = None) -> List[Ingredient]:
        """Search for the ingredients using their tier
        (:py:attr:`Ingredient.tier <corkus.objects.Ingredient.tier>`).

        :param tier: Tier of returned ingredients.
        :param timeout: Optionally override default timeout.
        """
        response = await self._request.get(
            version = APIVersion.V2,
            parameters = "ingredient/search/tier/" + str(tier),
            timeout = timeout
        )
        return [Ingredient(self._corkus, i) for i in response.get("data", [])]

    async def search_by_level(self, level: int, timeout: Optional[int] = None) -> List[Ingredient]:
        """Search for the ingredients using their level
        (:py:attr:`Ingredient.required_level <corkus.objects.Ingredient.required_level>`).

        :param level: Level of returned ingredients.
        :param timeout: Optionally override default timeout.
        """
        response = await self._request.get(
            version = APIVersion.V2,
            parameters = "ingredient/search/level/" + str(level),
            timeout = timeout
        )
        return [Ingredient(self._corkus, i) for i in response.get("data", [])]

    async def search_by_professions(self, symbol: LogicSymbol, professions: List[ProfessionType], timeout: Optional[int] = None) -> List[Ingredient]:
        """.. include:: moderate_route.rst

        Search for the ingredients using their required professions
        (:py:attr:`Ingredient.required_professions <corkus.objects.Ingredient.required_professions>`).

        :param symbol: Logic symbol to be used on ingredients list.
        :param professions: List of professions to search for.
        :param timeout: Optionally override default timeout.
        """
        professions = [s.value.lower() for s in professions]

        response = await self._request.get(
            version = APIVersion.V2,
            parameters = "ingredient/search/skills/" + symbol.value + ",".join(professions),
            timeout = timeout
        )
        return [Ingredient(self._corkus, i) for i in response.get("data", [])]

    async def search_by_sprite(self,
        symbol: LogicSymbol, *,
        id: Optional[int],
        damage: Optional[int] = None,
        timeout: Optional[int] = None
    ) -> List[Ingredient]:
        """.. include:: complex_route.rst

        Search for the ingredients using properties in
        :py:attr:`Ingredient.sprite <corkus.objects.Ingredient.sprite>`.

        :param symbol: Logic symbol to be used on this query.
        :param id: Ingredient property.
        :param damage: Ingredient property.
        :param timeout: Optionally override default timeout.
        """
        response = await self._request.get(
            version = APIVersion.V2,
            parameters = "ingredient/search/sprite/" + Utils.build_complex_query(symbol, id = id, damage = damage),
            timeout = timeout
        )
        return [Ingredient(self._corkus, i) for i in response.get("data", [])]

    async def search_by_identifications(self,
        symbol: LogicSymbol,
        identifications: List[Tuple[IdentificationType, Union[int, None], Union[int, None]]],
        timeout: Optional[int] = None
    ) -> List[Ingredient]:
        """.. include:: complex_route.rst

        Search for the ingredients using values of their identifications
        :py:attr:`Ingredient.identifications <corkus.objects.Ingredient.identifications>`.

        Example:

        .. code-block:: python

            # Search for ingredient that have xp bonus between 4 and 6 AND any loot bonus

            await corkus.ingredient.search_by_identifications(
                LogicSymbol.AND,
                [
                    (IdentificationType.XP_BONUS, 4, 6),
                    (IdentificationType.LOOT_BONUS, None, None),
                ]
            )

        :param symbol: Logic symbol to be used on this query.
        :param identifications:
            List of tuples with exactly three elements.

            * Type of the identification.

            * If a :py:class:`int`, the :py:attr:`min <corkus.objects.IdentificationValues.min>`
              for such identification must have that value. If :py:class:`None`, the
              :py:attr:`min <corkus.objects.IdentificationValues.min>` for such identification
              can be anything.

            * If a :py:class:`int`, the :py:attr:`max <corkus.objects.IdentificationValues.max>`
              for such identification must have that value. If :py:class:`None`, the
              :py:attr:`max <corkus.objects.IdentificationValues.max>` for such identification
              can be anything.

        :param timeout: Optionally override default timeout.
        """
        query = {}
        for i in identifications:
            query[IdentificationType.to_ingredient_api(i[0])] = (
                str(i[1] if i[1] is not None else "") + ";" + str(i[2] if i[2] is not None else "")
            )

        response = await self._request.get(
            version = APIVersion.V2,
            parameters = "ingredient/search/identifications/" + Utils.build_complex_query(symbol, **query),
            timeout = timeout
        )
        return [Ingredient(self._corkus, i) for i in response.get("data", [])]

    async def search_by_item_modifiers(self,
        symbol: LogicSymbol, *,
        durability: Optional[int] = None,
        strength: Optional[int] = None,
        dexterity: Optional[int] = None,
        intelligence: Optional[int] = None,
        defence: Optional[int] = None,
        agility: Optional[int] = None,
        timeout: Optional[int] = None
    ) -> List[Ingredient]:
        """.. include:: complex_route.rst

        Search for the ingredients using properties in
        :py:attr:`Ingredient.item_modifiers <corkus.objects.Ingredient.item_modifiers>`.

        :param symbol: Logic symbol to be used on this query.
        :param durability: Ingredient property.
        :param strength: Ingredient property.
        :param dexterity: Ingredient property.
        :param intelligence: Ingredient property.
        :param defence: Ingredient property.
        :param agility: Ingredient property.
        :param timeout: Optionally override default timeout.
        """
        response = await self._request.get(
            version = APIVersion.V2,
            parameters = "ingredient/search/itemOnlyIDs/" + Utils.build_complex_query(
                symbol,
                durability = durability,
                strength = strength,
                dexterity = dexterity,
                intelligence = intelligence,
                defence = defence,
                agility = agility
            ),
            timeout = timeout
        )
        return [Ingredient(self._corkus, i) for i in response.get("data", [])]

    async def search_by_consumable_modifiers(self,
        symbol: LogicSymbol, *,
        duration: Optional[int] = None,
        charges: Optional[int] = None,
        timeout: Optional[int] = None
    ) -> List[Ingredient]:
        """.. include:: complex_route.rst

        Search for the ingredients using properties in
        :py:attr:`Ingredient.consumable_modifiers <corkus.objects.Ingredient.consumable_modifiers>`.

        :param symbol: Logic symbol to be used on this query.
        :param duration: Ingredient property.
        :param charges: Ingredient property.
        :param timeout: Optionally override default timeout.
        """
        response = await self._request.get(
            version = APIVersion.V2,
            parameters = "ingredient/search/consumableOnlyIDs/" + Utils.build_complex_query(
                symbol,
                duration = duration,
                charges = charges
            ),
            timeout = timeout
        )
        return [Ingredient(self._corkus, i) for i in response.get("data", [])]



from __future__ import annotations
from typing import Union, TYPE_CHECKING
from .base import CorkusBase
from .enums import ProfessionType

if TYPE_CHECKING:
    from corkus import Corkus

class PlayerProfession(CorkusBase):
    """Information about :py:class:`Player` progress in a specified profession."""
    def __init__(self, corkus: Corkus, name: str, attributes: dict):
        super().__init__(corkus, attributes)
        self._name = name

    @property
    def name(self) -> str:
        """Pretty name of the profession like ``Alchemism``, ``Mining`` or even ``Combat``"""
        # pylint: disable=no-member
        return self.type.value.capitalize()

    @property
    def type(self) -> ProfessionType:
        """Type of the profession."""
        return ProfessionType(self._name.upper())

    @property
    def level(self) -> int:
        """Current level in profession. This value is minimum of ``1``
        and maximum of ``106`` for combat and ``132`` for other professions."""
        return self._attributes.get("level", 1)

    @property
    def level_progress(self) -> Union[float, None]:
        """Progress to next level in precentage 0-100%

        .. note::
            When a player reach max profession level ``132`` this will insteed
            return ``None`` but player can still collect XP.

            When a player reach combat level ``106`` this may exceed 100%.
        """
        return self._attributes.get("xp", 0)

    def __repr__(self) -> str:
        return f"<PlayerProfession name={self.name!r} level={self.level}>"



from __future__ import annotations
import time
import logging
import asyncio

logger = logging.getLogger("corkus.ratelimit")

class RateLimiter:
    def __init__(self) -> None:
        self._total = 180
        self._remaining = 180
        self._reset = 0

    @property
    def total(self) -> int:
        return self._total

    @property
    def remaining(self) -> int:
        if self.reset < 0:
            return self.total
        else:
            return self._remaining

    @property
    def reset(self) -> int:
        return self._reset - int(time.time())

    async def limit(self) -> None:
        if self.remaining <= 1:
            logger.info(f"You are being ratelimited, waiting for {self.reset}s")
            await asyncio.sleep(self.reset)

    def update(self, headers: dict) -> None:
        self._total = int(headers.get("ratelimit-limit", 180))
        self._remaining = int(headers.get("ratelimit-remaining", 180))
        self._reset = int(time.time()) + int(headers.get("ratelimit-reset", 0))



from __future__ import annotations
from typing import List, Literal, Union

from .base import CorkusBase
from .level_range import LevelRange
from .enums import ItemType, ProfessionType, ItemCategory
from .material import Material
from .identification_values import IdentificationValues

class Recipe(CorkusBase):
    """Needed materials and professions for `Crafting <https://wynncraft.fandom.com/wiki/Crafting>`_ a item with given properties."""
    @property
    def level(self) -> LevelRange:
        """Level range that this item should be used in."""
        return LevelRange(self._corkus, attributes = self._attributes.get("level", {}))

    @property
    def id(self) -> str:
        """Return the id of recipe like ``Food-13-15`` or ``Wand-63-65``."""
        return self._attributes.get("id", "")

    @property
    def type(self) -> ItemType:
        """Type of the item."""
        return ItemType(self._attributes.get("type", ItemType.WAND))

    @property
    def category(self) -> ItemCategory:
        """Category of the item."""
        return ItemCategory.from_type(self.type)

    @property
    def profession(self) -> ProfessionType:
        """Profession required to craft this item."""
        return ProfessionType(self._attributes.get("skill", ProfessionType.WOODWORKING))

    @property
    def materials(self) -> List[Material]:
        """Materials required to craft this item."""
        return [Material(self._corkus, m) for m in self._attributes.get("materials", [])]

    @property
    def health_or_damage(self) -> IdentificationValues:
        """If this item is a weapon this returns damage dealt by this item, if not, health bonus."""
        return IdentificationValues(self._corkus, attributes = self._attributes.get("healthOrDamage", {}))

    @property
    def duration(self) -> Union[None, IdentificationValues]:
        """If item is comsumbale returns how long it will work after use."""
        if "duration" in self._attributes:
            return IdentificationValues(self._corkus, attributes = self._attributes.get("duration", {}))
        else:
            return None

    @property
    def charges(self) -> Union[None, Literal[3]]:
        """If item is comsumbale returns how many times it can be used."""
        if self.type in (ItemType.POTION, ItemType.SCROLL, ItemType.FOOD):
            return 3
        else:
            return None

    @property
    def durability(self) -> Union[None, IdentificationValues]:
        """If item is armour accessory or weapon, return it's durability."""
        if "durability" in self._attributes:
            return IdentificationValues(self._corkus, attributes = self._attributes.get("durability", {}))
        else:
            return None

    def __repr__(self) -> str:
        return f"<Recipe level={self.level} type={self.type!r}>"



from __future__ import annotations
from typing import Optional

from .endpoint import Endpoint
from corkus.utils.request import APIVersion
from corkus.objects import SearchResult

class SearchEndpoint(Endpoint):
    async def guilds_and_players(self, term: str, timeout: Optional[int] = None) -> SearchResult:
        """Search for players and guilds in one requests.

        :param username: Search term.
        :param timeout: Optionally override default timeout.
        """
        response = await self._request.get(
            version = APIVersion.V1,
            parameters = "statsSearch&search=" + term,
            timeout = timeout
        )
        return SearchResult(self._corkus, response)



from __future__ import annotations
from .partial_base import PartialBase

class SkillPoints(PartialBase):
    """This is multi-purpose class for player skill points. You may see it as a reqired
    skill points for an item or skill points of the class."""

    def __init__(self, corkus, strength: int, dexterity: int, intelligence: int, defence: int, agility: int) -> None:
        super().__init__(corkus)
        self._strength = strength
        self._dexterity = dexterity
        self._intelligence = intelligence
        self._defence = defence
        self._agility = agility

    @property
    def strength(self) -> int:
        return self._strength

    @property
    def dexterity(self) -> int:
        return self._dexterity

    @property
    def intelligence(self) -> int:
        return self._intelligence

    @property
    def defence(self) -> int:
        return self._defence

    @property
    def agility(self) -> int:
        return self._agility

    def __repr__(self) -> str:
        return f"<SkillPoints strength={self.strength} dexterity={self.dexterity} intelligence={self.intelligence} defence={self.defence} agility={self.agility}>"



from __future__ import annotations
from typing import TYPE_CHECKING, Optional

from corkus.objects.partial_player import PartialPlayer

if TYPE_CHECKING:
    from corkus import Corkus
    from .uuid import CorkusUUID
    from .member import Member
    from .partial_guild import PartialGuild

class BasePartialMember(PartialPlayer):
    def __init__(self,
        corkus: Corkus ,
        uuid: CorkusUUID,
        username: str,
        guild: PartialGuild
    ):
        super().__init__(corkus, uuid=uuid, username=username)
        self._guild = guild

    @property
    def player(self) -> PartialPlayer:
        """Reduce to to :class:`PartialPlayer`. Usefull when you want
        to fetch player information not member information.
        """
        return super()

    @property
    def guild(self) -> PartialGuild:
        """The guild that this member is a part of."""
        return self._guild

    async def fetch(self, timeout: Optional[int] = None) -> Member:
        """Fetch more guild data about this member from API.

        .. include:: ../note_api_call.rst

        :param timeout: Optionally override default timeout.
        """
        guild = await self._guild.fetch(timeout)
        return next((m for m in guild.members if m.uuid == self._uuid), None)

    def __repr__(self) -> str:
        return f"<BasePartialMember username={self._username!r} guild={self._guild}>"



from __future__ import annotations
from .base import CorkusBase

class Quest(CorkusBase):
    """Represents a `Quest <https://wynncraft.fandom.com/wiki/Quests>`_ completed by a :py:class:`Player`"""
    @property
    def name(self) -> str:
        """The name of quest."""
        return self._attributes

    @property
    def mini(self) -> bool:
        """Is this quest a `Mini - Quest <https://wynncraft.fandom.com/wiki/Quests#Mini-Quests>`_."""
        return self.name.startswith("Mini-Quest - ")

    @property
    def wiki_url(self) -> str:
        """URL to the Wynncraft Wiki article about this quest."""
        if self.mini:
            return "https://wynncraft.fandom.com/wiki/Quests#Mini-Quests"
        else:
            return "https://wynncraft.fandom.com/wiki/" + self.name.replace(" ", "_")

    def __repr__(self) -> str:
        return f"<Quest name={self.name!r}>"



from __future__ import annotations
from .base import CorkusBase

class IngredientPositionModifiers(CorkusBase):
    """Modifier of :py:class:`Ingredient` that change effectiveness of other ingredients in the grid."""
    @property
    def right(self) -> int:
        """Ingredient effectiveness modifier in precentage to ingredients to the right of this one."""
        return self._attributes.get("right", 0)

    @property
    def left(self) -> int:
        """Ingredient effectiveness modifier in precentage to ingredients to the left of this one."""
        return self._attributes.get("left", 0)

    @property
    def above(self) -> int:
        """Ingredient effectiveness modifier in precentage to ingredients above of this one."""
        return self._attributes.get("above", 0)

    @property
    def under(self) -> int:
        """Ingredient effectiveness modifier in precentage to ingredients under this one."""
        return self._attributes.get("under", 0)

    @property
    def touching(self) -> int:
        """Ingredient effectiveness modifier in precentage to ingredients touching this one."""
        return self._attributes.get("touching", 0)

    @property
    def not_touching(self) -> int:
        """Ingredient effectiveness modifier in precentage to ingredients not touching this one."""
        return self._attributes.get("notTouching", 0)



from __future__ import annotations
from typing import TYPE_CHECKING, Optional, Literal

from .partial_player import PartialPlayer

if TYPE_CHECKING:
    from corkus import Corkus
    from .uuid import CorkusUUID
    from .server import Server


class PartialOnlinePlayer(PartialPlayer):
    """Represents a :py:class:`PartialPlayer` that is currently online."""
    def __init__(self, corkus: Corkus, server: Server, *, uuid: Optional[CorkusUUID] = None, username: Optional[str] = None):
        super().__init__(corkus, uuid = uuid, username = username)
        self._server = server

    @property
    def online(self) -> Literal[True]:
        """Is player online right now, always ``True`` obviously."""
        return True

    @property
    def server(self) -> Server:
        """The server that this player is currently on."""
        return self._server

    def __repr__(self) -> str:
        return super().__repr__().replace("PartialPlayer", f"PartialOnlinePlayer server={self._server.name!r} ")



from __future__ import annotations
from os import error
from typing import Optional
from enum import Enum
import asyncio
from aiohttp.client import ClientSession, ClientResponse
import copy
import logging

from .cache import CorkusCache
from .ratelimit import RateLimiter
from corkus.version import __version__
from corkus.errors import BadRequest, WynncraftServerError, RatelimitExceeded, HTTPError, CorkusTimeoutError

logger = logging.getLogger("corkus.request")

class APIVersion(Enum):
    V1 = "https://api.wynncraft.com/public_api.php?action="
    V2 = "https://api.wynncraft.com/v2/"

class CorkusRequest:
    """CorkusRequest is a internal overlay over aiohttp to simplify API calls."""

    def __init__(self,
        timeout: Optional[int] = 0,
        api_key: Optional[str] = None,
        session: Optional[ClientSession] = None,
        ratelimit_enable: Optional[bool] = True,
        cache_enable: Optional[bool] = True,
    ) -> None:
        self.ratelimit = RateLimiter()
        self.cache = CorkusCache()
        self.ratelimit_enable = ratelimit_enable
        self.cache_enable = cache_enable

        self._session = session
        self.timeout = timeout
        if session is None:
            headers = {
                "User-Agent": f"Corkus.py/{__version__}",
                "Content-Type": "application/json"
            }
            if api_key is not None:
                headers["apikey"] = api_key
            self._session = ClientSession(headers = headers)

    async def get(self, version: APIVersion, parameters: str, timeout: Optional[int]) -> dict:
        url = version.value + parameters

        if self.cache_enable:
            cache_element = self.cache.get(url)
            if cache_element:
                return copy.copy(cache_element.content)

        if self.ratelimit_enable:
            await self.ratelimit.limit()

        try:
            response = await self._session.get(url, timeout = timeout or self.timeout)
            data = await response.json()
        except asyncio.TimeoutError:
            raise CorkusTimeoutError(timeout, error)

        self.ratelimit.update(response.headers)
        self._fix_status_codes(data, response)

        if 200 <= response.status < 400:
            self.cache.add(url, response.headers, data)
            return copy.copy(data)

        elif response.status >= 500:
            raise WynncraftServerError(response)
        elif response.status >= 429:
            raise RatelimitExceeded(response)
        elif response.status >= 400:
            raise BadRequest(response)
        else:
            raise HTTPError(response)

    async def close(self) -> None:
        return await self._session.close()

    def _fix_status_codes(self, data: dict, response: ClientResponse) -> None:
        """Fix endpoints that return wrong status codes."""

        # https://github.com/Wynncraft/WynncraftAPI/issues/63
        # and other similar
        if data.get("error") is not None and response.status != 400:
            logger.debug(f"Got error field but status code is {response.status}- changing to: 400")
            response.status = 400

        elif data.get("status", 200) != response.status:
            status_field = data.get("status", 200)
            new_status = max(data.get("status", 200), response.status)
            logger.debug(f"Status field [{status_field}] is diffrent than status code [{response.status}] - changing to: {new_status}")
            response.status = new_status



from __future__ import annotations
from typing import TYPE_CHECKING

from corkus.objects.base_partial_member import BasePartialMember

if TYPE_CHECKING:
    from corkus import Corkus
    from .uuid import CorkusUUID
    from .partial_guild import PartialGuild

class LeaderboardPartialMember(BasePartialMember):
    """Diffrent version of :py:class:`PartialMember` returned by :py:class:`LeaderboardEndpoint <corkus.endpoints.LeaderboardEndpoint>`."""
    def __init__(self,
        corkus: Corkus,
        uuid: CorkusUUID,
        username: str,
        guild: PartialGuild,
        tag: str
        ):
        super().__init__(corkus, uuid, username, guild)
        self._tag = tag

    @property
    def tag(self) -> str:
        """Prefix displayed as a shorter 3 or 4 letters representation of guild.
        Like ``AVO``, ``ERN`` or ``YIN``. It's mostly upper-case but it's not a requirement.
        It's unique across all guild."""
        return self._tag

    def __repr__(self) -> str:
        return f"<LeaderboardPartialMember username={self._username!r} guild={self._guild}>"



# Scrapped from: https://www.wynndata.tk/items/

major_ids = {
    "HERO": {
        "name": "Saviour's Sacrifice",
        "effect": "While under 50% maximum health, nearby allies gain 50% bonus damage and defence"
    },
    "MAGNET": {
        "name": "Magnet",
        "effect": "Pulls items within an 8 block radius towards you"
    },
    "LIGHTWEIGHT": {
        "name": "Lightweight",
        "effect": "You no longer take fall damage"
    },
    "ENTROPY": {
        "name": "Entropy",
        "effect": "Meteor falls three times faster"
    },
    "MADNESS": {
        "name": "Madness",
        "effect": "Casts a random ability every 3 seconds"
    },
    "EXPLOSIVE_IMPACT": {
        "name": "Explosive Impact",
        "effect": "Your \"Exploding\" ID can trigger whet hitting mobs with your basic attack"
    },
    "ARCANES": {
        "name": "Transcendence",
        "effect": "50% chance for spells to cost no mana when casted"
    },
    "PLAGUE": {
        "name": "Plague",
        "effect": "Poisoned mobs spread their poison to nearby mobs"
    },
    "RALLY": {
        "name": "Rally",
        "effect": "Charge heals you by 10% and nearby allies by 15% on impact, but becomes harmless"
    },
    "FISSION": {
        "name": "Fission",
        "effect": "Explosions from your \"Exploding\" ID are twice as big and twice as strong"
    },
    "TAUNT": {
        "name": "Taunt",
        "effect": "Mobs within 12 blocks target you upon casting War Scream"
    },
    "HAWKEYE": {
        "name": "Hawkeye",
        "effect": "Arrow storm fires 5 arrows each dealing 80% damage"
    },
    "ROVINGASSASSIN": {
        "name": "Roving Assassin",
        "effect": "Vanish no longer drains mana while invisible"
    },
    "ALTRUISM": {
        "name": "Heart of the Pack",
        "effect": "Nearby players gain 35% of the health you naturally regenerate"
    },
    "FLASHFREEZE": {
        "name": "Flashfreeze",
        "effect": "Ice Snake is instant and freezes for longer at a reduced range"
    },
    "FREERUNNER": {
        "name": "Freerunner",
        "effect": "Double your sprint speed when your sprint bar is under 30%"
    },
    "GREED": {
        "name": "Greed",
        "effect": "Picking up emeralds heals you and nearby players for 15% max health"
    },
    "CAVALRYMAN": {
        "name": "Cavalryman",
        "effect": "You may cast spells and attack with a 70% damage penalty while on a horse"
    },
    "SORCERY": {
        "name": "Sorcery",
        "effect": "30% chance for spells and attacks to cast a second time at no additional cost"
    },
    "GUARDIAN": {
        "name": "Guardian",
        "effect": "50% of damage taken by nearby allies is redirected to you"
    },
    "PEACEFUL_EFFIGY": {
        "name": "Peaceful Effigy",
        "effect": "Your totem will last twice as long"
    },
    "CHERRY_BOMBS": {
        "name": "Cherry Bombs",
        "effect": "Your Smoke Bombs explodes instantly on contact, dealing 110% damage each"
    },
    "FURIOUS_EFFIGY": {
        "name": "Furious Effigy",
        "effect": "Totem effects are twice as fast, but duration is halved"
    }
}



from __future__ import annotations
from typing import TYPE_CHECKING

from .base import CorkusBase
from corkus.data.major_ids import major_ids

if TYPE_CHECKING:
    from corkus import Corkus

class MajorIdentification(CorkusBase):
    """Major IDs are a type of special :py:class:`Identification` but, they are not
    randomly rolled and they provide a wide range of effects that cannot be acquired
    elsewhere"""
    @property
    def id(self) -> str:
        """The id of the Major ID like ``PLAGUE`` or ``ARCANES``."""
        return self._attributes

    @property
    def name(self) -> str:
        """The name of the Major ID like ``Plague`` or ``Transcendence``."""
        return self._id_data.get("name", "")

    @property
    def effect(self) -> str:
        """The description of effect this Major ID provide."""
        return self._id_data.get("effect", "")

    @property
    def _id_data(self) -> dict:
        return major_ids.get(self.id, {})

    def __repr__(self) -> str:
        return f"<MajorIdentification name={self.name!r}>"



from __future__ import annotations
from typing import List, Optional

from .endpoint import Endpoint
from corkus.utils.request import APIVersion
from corkus.objects import LeaderboardPlayer, LeaderboardGuild, Timeframe

class LeaderboardEndpoint(Endpoint):
    async def guild(self, timeout: Optional[int] = None) -> List[LeaderboardGuild]:
        """Returns a leaderboard of guilds with a length of 100. Sorted by territories
        currently owned, then level.

        :param timeout: Optionally override default timeout."""
        response = await self._request.get(
            version = APIVersion.V1,
            parameters = "statsLeaderboard&type=guild&timeframe=alltime",
            timeout = timeout
        )
        return [LeaderboardGuild(self._corkus, g) for g in response.get("data", [])]

    async def combat(self, timeout: Optional[int] = None) -> List[LeaderboardPlayer]:
        """Returns a leaderboard of players with a length of 100. Sorted by player combat level/xp.

        :param timeout: Optionally override default timeout.
        """
        response = await self._request.get(
            version = APIVersion.V1,
            parameters = "statsLeaderboard&type=player&timeframe=alltime",
            timeout = timeout
        )
        return [LeaderboardPlayer(self._corkus, g) for g in response.get("data", [])]

    async def pvp(self, *, timeframe: Optional[Timeframe] = Timeframe.ALL_TIME, timeout: Optional[int] = None) -> List[LeaderboardPlayer]:
        """Returns a leaderboard of players with a length of 100. Sorted by
        :py:attr:`LeaderboardPlayer.pvp_kills <corkus.objects.LeaderboardPlayer.pvp_kills>`.

        .. caution::
            Since `The Nether <https://wynncraft.fandom.com/wiki/The_Nether>`_ is disabled
            changing ``timeframe`` doesn't do anything.

        :param timeframe: Timeframe for results will be returned.
        :param timeout: Optionally override default timeout.
        """
        response = await self._request.get(
            version = APIVersion.V1,
            parameters = "statsLeaderboard&type=pvp&timeframe=" + timeframe.value,
            timeout = timeout
        )
        return [LeaderboardPlayer(self._corkus, g) for g in response.get("data", [])]



from __future__ import annotations
from typing import TYPE_CHECKING, Union, Optional

from .partial_base import PartialBase
from .server import ServerType

if TYPE_CHECKING:
    from corkus import Corkus
    from .server import Server

class PartialServer(PartialBase):
    """Represents a ``Partial`` version of :py:class:`Server`."""
    def __init__(self, corkus: Corkus, name: str):
        super().__init__(corkus)
        self._name = name

    @property
    def name(self) -> str:
        """The name of server."""
        return self._name

    @property
    def type(self) -> ServerType:
        """Type of the server, most servers are stanard servers."""
        return ServerType("".join([i for i in self._name if not i.isdigit()]))

    @property
    def regular(self) -> bool:
        """Is this server a regular server (:py:attr:`type` is :py:attr:`ServerType.REGULAR`)."""
        return self.type == ServerType.REGULAR

    async def fetch(self, timeout: Optional[int] = None) -> Union[Server, None]:
        """Fetch full server information from API. Returns ``None`` if server no longer exist.

        .. include:: ../note_api_call.rst

        :param timeout: Optionally override default timeout.
        """
        online_players = await self._corkus.network.online_players(timeout)
        for server in online_players.servers:
            if server.name == self.name:
                return server
        return None

    def __repr__(self) -> str:
        return f"<PartialServer name={self.name!r}>"



from __future__ import annotations
from typing import List, Optional

from .endpoint import Endpoint
from corkus.utils.request import APIVersion
from corkus.objects import Recipe, PartialRecipe, ItemType, ProfessionType, LogicSymbol
from corkus.utils.utils import Utils

class RecipeEndpoint(Endpoint):
    async def list_all(self, timeout: Optional[int] = None) -> List[PartialRecipe]:
        """List all available recipes.

        :param timeout: Optionally override default timeout.
        """
        response = await self._request.get(
            version = APIVersion.V2,
            parameters = "recipe/list",
            timeout = timeout
        )
        return [PartialRecipe(self._corkus, r) for r in response.get("data", [])]

    async def get(self) -> Recipe:
        """.. include:: ../note_not_implemented.rst"""
        raise NotImplementedError

    async def get_by_id(self, recipe_id: str, timeout: Optional[int] = None) -> Recipe:
        """Get information about recipe by id like ``Food-13-15`` or ``Wand-63-65``.

        :param timeout: Optionally override default timeout.
        """
        response = await self._request.get(
            version = APIVersion.V2,
            parameters = "recipe/get/" + recipe_id,
            timeout = timeout
        )
        return Recipe(self._corkus, response.get("data", {})[0])

    async def search_by_type(self, type: ItemType, timeout: Optional[int] = None) -> List[Recipe]:
        """Search for the ingredients using type
        (:py:attr:`Recipe.type <corkus.objects.Recipe.type>`).

        :param name: Type of returned recipes.
        :param timeout: Optionally override default timeout.
        """
        response = await self._request.get(
            version = APIVersion.V2,
            parameters = "recipe/search/type/" + type.value.lower(),
            timeout = timeout
        )
        return [Recipe(self._corkus, r) for r in response.get("data", [])]

    async def search_by_profession(self, profession: ProfessionType, timeout: Optional[int] = None) -> List[Recipe]:
        """Search for the ingredients using profession
        (:py:attr:`Recipe.profession <corkus.objects.Recipe.profession>`).

        :param profession: Required profession for returned recipes.
        :param timeout: Optionally override default timeout.
        """
        response = await self._request.get(
            version = APIVersion.V2,
            parameters = "recipe/search/skill/" + profession.value.lower(),
            timeout = timeout
        )
        return [Recipe(self._corkus, r) for r in response.get("data", [])]

    async def search_by_level(self,
        symbol: LogicSymbol, *,
        min: Optional[int] = None,
        max: Optional[int] = None,
        timeout: Optional[int] = None
    ) -> List[Recipe]:
        """.. include:: complex_route.rst

        Search for the recipes using their
        :py:attr:`Recipe.level <corkus.objects.Recipe.level>`.

        :param symbol: Logic symbol to be used on this query.
        :param min: Minimum level.
        :param max: Maximum level.
        :param timeout: Optionally override default timeout.
        """
        response = await self._request.get(
            version = APIVersion.V2,
            parameters = "recipe/search/level/" + Utils.build_complex_query(symbol, minimum = min, maximum = max),
            timeout = timeout
        )
        return [Recipe(self._corkus, i) for i in response.get("data", [])]

    async def search_by_durability(self,
        symbol: LogicSymbol, *,
        min: Optional[int] = None,
        max: Optional[int] = None,
        timeout: Optional[int] = None
    ) -> List[Recipe]:
        """.. include:: complex_route.rst

        Search for the recipes using their
        :py:attr:`Recipe.durability <corkus.objects.Recipe.durability>`.

        :param symbol: Logic symbol to be used on this query.
        :param min: Minimum durability identification.
        :param max: Maximum durability identification.
        :param timeout: Optionally override default timeout.
        """
        response = await self._request.get(
            version = APIVersion.V2,
            parameters = "recipe/search/durability/" + Utils.build_complex_query(symbol, minimum = min, maximum = max),
            timeout = timeout
        )
        return [Recipe(self._corkus, i) for i in response.get("data", [])]

    async def search_by_health_or_damage(self,
        symbol: LogicSymbol, *,
        min: Optional[int] = None,
        max: Optional[int] = None,
        timeout: Optional[int] = None
    ) -> List[Recipe]:
        """.. include:: complex_route.rst

        Search for the recipes using their
        :py:attr:`Recipe.health_or_damage <corkus.objects.Recipe.health_or_damage>`.

        :param symbol: Logic symbol to be used on this query.
        :param min: Minimum *health or damage* identification.
        :param max: Maximum *health or damage* identification.
        :param timeout: Optionally override default timeout.
        """
        response = await self._request.get(
            version = APIVersion.V2,
            parameters = "recipe/search/healthOrDamage/" + Utils.build_complex_query(symbol, minimum = min, maximum = max),
            timeout = timeout
        )
        return [Recipe(self._corkus, i) for i in response.get("data", [])]

    async def search_by_duration(self,
        symbol: LogicSymbol, *,
        min: Optional[int] = None,
        max: Optional[int] = None,
        timeout: Optional[int] = None
    ) -> List[Recipe]:
        """.. include:: complex_route.rst

        Search for the recipes using their
        :py:attr:`Recipe.duration <corkus.objects.Recipe.duration>`.

        :param symbol: Logic symbol to be used on this query.
        :param min: Minimum duration identification.
        :param max: Maximum duration identification.
        :param timeout: Optionally override default timeout.
        """
        response = await self._request.get(
            version = APIVersion.V2,
            parameters = "recipe/search/duration/" + Utils.build_complex_query(symbol, minimum = min, maximum = max),
            timeout = timeout
        )
        return [Recipe(self._corkus, i) for i in response.get("data", [])]



from __future__ import annotations
from corkus.objects.base_player import PlayerTag
from datetime import datetime, timezone
from enum import Enum
from typing import TYPE_CHECKING, List, Tuple

from .base import CorkusBase
from .player_profession import PlayerProfession
from .quest import Quest
from .player_statistics import ClassStatistics
from .player_gamemodes import HardcoreType, PlayerGamemodes
from .dungeon import Dungeon, DungeonType
from .raid import Raid
from .enums import ProfessionType
from .skill_points import SkillPoints
from .playtime import PlayerPlaytime

if TYPE_CHECKING:
    from corkus import Corkus
    from .player import Player

class ClassType(Enum):
    """Type of :py:class:`PlayerClass`."""
    ARCHER = "ARCHER"
    WARRIOR = "WARRIOR"
    MAGE = "MAGE"
    ASSASSIN = "ASSASSIN"
    SHAMAN = "SHAMMAN"

    HUNTER = "HUNTER"
    KNIGHT = "KNIGHT"
    DARK_WIZARD = "DARK_WIZARD"
    NINJA = "NINJA"
    SKYSEER = "SKYSEER"

class PlayerClass(CorkusBase):
    """Represents one of the `player classes <https://wynncraft.fandom.com/wiki/Classes>`_
    that a :py:class:`Player` have currently active."""
    def __init__(self, corkus: Corkus, player: Player, attributes: dict):
        self._player = player
        super().__init__(corkus, attributes)

    @property
    def name(self) -> str:
        """The official name of the class for example: ``mage``, ``knight1`` or ``darkwizard3``."""
        return self._attributes.get("name", "")

    @property
    def display_name(self) -> str:
        """Pretty name to display to end-user like ``Archer``, ``Mage`` or ``Dark Wizard``."""
        return self.type.value.replace("_"," ").title()

    @property
    def approximate_create(self) -> Tuple[datetime, datetime]:
        """Approximate time when class was created.

        Wynncraft API does not provide information
        when a class was created but this property makes a educated guess and provide two datetimes
        between which this class was created.
        """
        SE_UPDATE = datetime(2019, 12, 8, tzinfo = timezone.utc)
        ECONOMY_UPDATE_PATCH1 = datetime(2019, 6, 8, tzinfo = timezone.utc)
        ECONOMY_UPDATE = datetime(2019, 1, 18, tzinfo = timezone.utc)
        DUNGEONS_UPDATE = datetime(2017, 12, 15, tzinfo = timezone.utc)
        GAMEPLAY_UPDATE = datetime(2016, 8, 28, tzinfo = timezone.utc)
        MOB_UPDATE = datetime(2014, 8, 1, tzinfo = timezone.utc)

        start = self._player.join_date
        end = self._player.last_online

        if self.kind == ClassType.SHAMAN:
            start = max(SE_UPDATE, start)

        if self.reskinned and self._player.tag not in (PlayerTag.HERO, PlayerTag.CHAMPION):
            start = max(SE_UPDATE, start)

        if self.pre_economy_update or self.statistics.pvp_deaths > 0 or self.statistics.pvp_kills > 0:
            end = min(ECONOMY_UPDATE, end)

        if self.gamemode.craftsman or self.gamemode.hardcore != HardcoreType.DISABLED or self.gamemode.ironman:
            start = max(ECONOMY_UPDATE_PATCH1, start)

        if self.gamemode.hunted:
            start = max(SE_UPDATE, start)

        if any(d.type == DungeonType.REMOVED_MINI for d in self.dungeons):
            end = min(DUNGEONS_UPDATE, end)

        if any(d.type == DungeonType.REMOVED for d in self.dungeons):
            end = min(GAMEPLAY_UPDATE, end)

        if self.statistics.swarms_won > 0:
            end = min(MOB_UPDATE, end)

        return (start, end)

    @property
    def kind(self) -> ClassType:
        """Class type ignoring re-skinned variants. :py:attr:`ClassType.DARK_WIZARD` is converted to :py:attr:`ClassType.MAGE` etc."""
        class_type = self.type
        if class_type == ClassType.HUNTER:
            return ClassType.ARCHER
        elif class_type == ClassType.KNIGHT:
            return ClassType.WARRIOR
        elif class_type == ClassType.DARK_WIZARD:
            return ClassType.MAGE
        elif class_type == ClassType.NINJA:
            return ClassType.ASSASSIN
        elif class_type == ClassType.SKYSEER:
            return ClassType.SHAMAN
        else:
            return class_type

    @property
    def reskinned(self) -> bool:
        """Does this class use re-skinned variant."""
        return self.type != self.kind

    @property
    def type(self) -> ClassType:
        """Class type including re-skinned variants."""
        name = "".join([i for i in self.name if not i.isdigit()])
        if name == "darkwizard":
            name = "dark wizard"
        name = name.upper().replace(" ", "_")
        return ClassType(name)

    @property
    def quests(self) -> List[Quest]:
        """List of all quests completed by player on this class."""
        return [Quest(self._corkus, q) for q in self._attributes.get("quests", {}).get("list", [])]

    @property
    def playtime(self) -> PlayerPlaytime:
        """Time that player spent on wynncraft servers using this class."""
        return PlayerPlaytime(self._attributes.get("playtime", 0))

    @property
    def combined_level(self) -> int:
        """Combined level of all professions including combat."""
        return self._attributes.get("level", 0)

    @property
    def statistics(self) -> ClassStatistics:
        """General statistics for this class."""
        return ClassStatistics(self._corkus, self._attributes)

    @property
    def skill_points(self) -> SkillPoints:
        """Skill points of this class."""
        skills = self._attributes.get("skills", {})
        return SkillPoints(
            corkus = self._corkus,
            strength = skills.get("strength", 0),
            dexterity = skills.get("dexterity", 0),
            intelligence = skills.get("intelligence", 0),
            defence = skills.get("defence", 0),
            agility = skills.get("agility", 0)
        )

    @property
    def dungeons(self) -> List[Dungeon]:
        """List of dungeons completed by this class."""
        return [Dungeon(self._corkus, d) for d in self._attributes.get("dungeons", {}).get("list", [])]

    @property
    def raids(self) -> List[Raid]:
        """List of raids completed by this class."""
        return [Raid(self._corkus, d) for d in self._attributes.get("raids", {}).get("list", [])]

    @property
    def gamemode(self) -> PlayerGamemodes:
        """Challenge gamemodes that are enabled on this class."""
        return PlayerGamemodes(self._corkus, self._attributes.get("gamemode", {}), self.statistics.deaths)

    @property
    def professions(self) -> List[PlayerProfession]:
        """Returns a list of all players professions."""
        return [PlayerProfession(self._corkus, name, attr) for name, attr in self._attributes.get("professions", {}).items()]

    @property
    def combat(self) -> PlayerProfession:
        """Shortcut to combat profession."""
        return self.get_profession(ProfessionType.COMBAT)

    @property
    def pre_economy_update(self) -> str:
        """Did this class achieve level 101 before `1.18 Economy Update
        <https://forums.wynncraft.com/threads/1-18-the-economy-update-changelog.238340/>`_
        when level cap was risen.
        """
        return self._attributes.get("preEconomyUpdate")

    def get_profession(self, profession: ProfessionType) -> PlayerProfession:
        """Returns a profession with the given profession type.

        :param profession: Type of profession."""
        return next((p for p in self.professions if p.type == profession), None)

    def __repr__(self) -> str:
        return f"<PlayerClass type={self.display_name!r}>"



# Translation of identification values returned by API
# to values used by library

from corkus.objects.identification_type import IdentificationType

ids_convert = [
    {
        "type": IdentificationType.AIR_DEFENSE,
        "items_api": "airDefense",
        "ingredient_api": ""
    }, {
        "type": IdentificationType.EARTH_DEFENSE,
        "items_api": "earthDefense",
        "ingredient_api": ""
    }, {
        "type": IdentificationType.FIRE_DEFENSE,
        "items_api": "fireDefense",
        "ingredient_api": ""
    }, {
        "type": IdentificationType.THUNDER_DEFENSE,
        "items_api": "thunderDefense",
        "ingredient_api": ""
    }, {
        "type": IdentificationType.WATER_DEFENSE,
        "items_api": "waterDefense",
        "ingredient_api": ""
    }, {
        "type": IdentificationType.AIR_DEFENSE_BONUS,
        "items_api": "bonusAirDefense",
        "ingredient_api": "AIRDEFENSE"
    }, {
        "type": IdentificationType.EARTH_DEFENSE_BONUS,
        "items_api": "bonusEarthDefense",
        "ingredient_api": "EARTHDEFENSE"
    }, {
        "type": IdentificationType.FIRE_DEFENSE_BONUS,
        "items_api": "bonusFireDefense",
        "ingredient_api": "FIREDEFENSE"
    }, {
        "type": IdentificationType.THUNDER_DEFENSE_BONUS,
        "items_api": "bonusThunderDefense",
        "ingredient_api": "THUNDERDEFENSE"
    }, {
        "type": IdentificationType.WATER_DEFENSE_BONUS,
        "items_api": "bonusWaterDefense",
        "ingredient_api": "WATERDEFENSE"
    }, {
        "type": IdentificationType.DAMAGE_BONUS,
        "items_api": "damageBonus",
        "ingredient_api": "DAMAGEBONUS"
    }, {
        "type": IdentificationType.DAMAGE_BONUS_RAW,
        "items_api": "damageBonusRaw",
        "ingredient_api": "DAMAGEBONUSRAW"
    }, {
        "type": IdentificationType.SPELL_DAMAGE,
        "items_api": "spellDamage",
        "ingredient_api": "SPELLDAMAGE"
    }, {
        "type": IdentificationType.SPELL_DAMAGE_RAW,
        "items_api": "spellDamageRaw",
        "ingredient_api": "SPELLDAMAGERAW"
    }, {
        "type": IdentificationType.AIR_DAMAGE_BONUS,
        "items_api": "bonusAirDamage",
        "ingredient_api": "AIRDAMAGEBONUS"
    }, {
        "type": IdentificationType.EARTH_DAMAGE_BONUS,
        "items_api": "bonusEarthDamage",
        "ingredient_api": "EARTHDAMAGEBONUS"
    }, {
        "type": IdentificationType.FIRE_DAMAGE_BONUS,
        "items_api": "bonusFireDamage",
        "ingredient_api": "FIREDAMAGEBONUS"
    }, {
        "type": IdentificationType.THUNDER_DAMAGE_BONUS,
        "items_api": "bonusThunderDamage",
        "ingredient_api": "THUNDERDAMAGEBONUS"
    }, {
        "type": IdentificationType.WATER_DAMAGE_BONUS,
        "items_api": "bonusWaterDamage",
        "ingredient_api": "WATERDAMAGEBONUS"
    }, {
        "type": IdentificationType.AGILITY_POINTS,
        "items_api": "agilityPoints",
        "ingredient_api": "AGILITYPOINTS"
    }, {
        "type": IdentificationType.DEFENSE_POINTS,
        "items_api": "defensePoints",
        "ingredient_api": "DEFENSEPOINTS"
    }, {
        "type": IdentificationType.DEXTERITY_POINTS,
        "items_api": "dexterityPoints",
        "ingredient_api": "DEXTERITYPOINTS"
    }, {
        "type": IdentificationType.INTELLIGENCE_POINTS,
        "items_api": "intelligencePoints",
        "ingredient_api": "INTELLIGENCEPOINTS"
    }, {
        "type": IdentificationType.STRENGTH_POINTS,
        "items_api": "strengthPoints",
        "ingredient_api": "STRENGTHPOINTS"
    }, {
        "type": IdentificationType.POISON,
        "items_api": "poison",
        "ingredient_api": "POISON"
    }, {
        "type": IdentificationType.MANA_STEAL,
        "items_api": "manaSteal",
        "ingredient_api": "MANASTEAL"
    }, {
        "type": IdentificationType.MANA_REGEN,
        "items_api": "manaRegen",
        "ingredient_api": "MANAREGEN"
    }, {
        "type": IdentificationType.WALK_SPEED,
        "items_api": "speed",
        "ingredient_api": "SPEED"
    }, {
        "type": IdentificationType.HEALTH_BONUS,
        "items_api": "healthBonus",
        "ingredient_api": "HEALTHBONUS"
    }, {
        "type": IdentificationType.HEALTH_REGEN,
        "items_api": "healthRegen",
        "ingredient_api": "HEALTHREGEN"
    }, {
        "type": IdentificationType.HEALTH_REGEN_RAW,
        "items_api": "healthRegenRaw",
        "ingredient_api": "HEALTHREGENRAW"
    }, {
        "type": IdentificationType.ATTACK_SPEED,
        "items_api": "attackSpeedBonus",
        "ingredient_api": "ATTACKSPEED"
    }, {
        "type": IdentificationType.LIFE_STEAL,
        "items_api": "lifeSteal",
        "ingredient_api": "LIFESTEAL"
    }, {
        "type": IdentificationType.REFLECTION,
        "items_api": "reflection",
        "ingredient_api": "REFLECTION"
    }, {
        "type": IdentificationType.THORNS,
        "items_api": "thorns",
        "ingredient_api": "THORNS"
    }, {
        "type": IdentificationType.EXPLODING,
        "items_api": "exploding",
        "ingredient_api": "EXPLODING"
    }, {
        "type": IdentificationType.LOOT_BONUS,
        "items_api": "lootBonus",
        "ingredient_api": "LOOTBONUS"
    }, {
        "type": IdentificationType.XP_BONUS,
        "items_api": "xpBonus",
        "ingredient_api": "XPBONUS"
    }, {
        "type": IdentificationType.EMERALD_STEALING,
        "items_api": "emeraldStealing",
        "ingredient_api": "EMERALDSTEALING"
    }, {
        "type": IdentificationType.GATHER_XP_BONUS,
        "items_api": "gatherXpBonus",
        "ingredient_api": "GATHER_XP_BONUS"
    }, {
        "type": IdentificationType.GATHER_SPEED,
        "items_api": "gatherSpeed",
        "ingredient_api": "GATHER_SPEED"
    }, {
        "type": IdentificationType.SOUL_POINTS,
        "items_api": "soulPoints",
        "ingredient_api": "SOULPOINTS"
    }, {
        "type": IdentificationType.SPRINT,
        "items_api": "sprintRegen",
        "ingredient_api": ""
    }, {
        "type": IdentificationType.SPRINT_REGEN,
        "items_api": "sprint",
        "ingredient_api": ""
    }, {
        "type": IdentificationType.JUMP_HEIGHT,
        "items_api": "jumpHeight",
        "ingredient_api": ""
    }, {
        "type": IdentificationType.LOOT_QUALITY,
        "items_api": "lootQuality",
        "ingredient_api": ""
    }, {
        "type": IdentificationType.RAINBOW_SPELL_DAMAGE_RAW,
        "items_api": "rainbowSpellDamageRaw",
        "ingredient_api": ""
    }, {
        "type": IdentificationType.SPELL_COST_1,
        "items_api": "spellCostRaw1",
        "ingredient_api": ""
    }, {
        "type": IdentificationType.SPELL_COST_1_RAW,
        "items_api": "spellCostPct1",
        "ingredient_api": ""
    }, {
        "type": IdentificationType.SPELL_COST_2,
        "items_api": "spellCostRaw2",
        "ingredient_api": ""
    }, {
        "type": IdentificationType.SPELL_COST_2_RAW,
        "items_api": "spellCostPct2",
        "ingredient_api": ""
    }, {
        "type": IdentificationType.SPELL_COST_3,
        "items_api": "spellCostRaw3",
        "ingredient_api": ""
    }, {
        "type": IdentificationType.SPELL_COST_3_RAW,
        "items_api": "spellCostPct3",
        "ingredient_api": ""
    }, {
        "type": IdentificationType.SPELL_COST_4,
        "items_api": "spellCostRaw5",
        "ingredient_api": ""
    }, {
        "type": IdentificationType.SPELL_COST_4_RAW,
        "items_api": "spellCostPct4",
        "ingredient_api": ""
    }
]



from __future__ import annotations
from .base import CorkusBase
from .partial_teritories import PartialTeritories
from datetime import datetime
from iso8601 import iso8601
from .guild_banner import GuildBanner

class BaseGuild(CorkusBase):
    @property
    def name(self) -> str:
        """The name of the guild."""
        return self._attributes.get("name", "")

    @property
    def tag(self) -> str:
        """Prefix displayed as a shorter 3 or 4 letters representation of guild.
        Like ``AVO``, ``ERN`` or ``YIN``. It's mostly upper-case but it's not a requirement.
        It's unique across all guild.
        """
        return self._attributes.get("prefix", "")

    @property
    def created(self) -> datetime:
        """Datetime when guild was created."""
        return iso8601.parse_date(self._attributes.get("created", "1970"))

    @property
    def level(self) -> int:
        """Level of the guild from ``1`` to ``100``. Guild members can use ``/guild xp`` command
        to contribute XP to the guild. Leveling guild grants perks like more
        member slots or guild bank upgrades. The exact XP values are available
        `on the wiki <https://wynncraft.fandom.com/wiki/Guilds#Leveling_Guild>`_.
        """
        return self._attributes.get("level", 1)

    @property
    def territories(self) -> PartialTeritories:
        """Teritories owned by the guild throughout wars."""
        return PartialTeritories(
            corkus = self._corkus,
            guild = self,
            count = self._attributes.get("territories", 0)
        )

    @property
    def banner(self) -> GuildBanner:
        """Information about appearance of guild banner and it's tier."""
        return GuildBanner(self._corkus, self._attributes.get("banner", {}))

    def __repr__(self) -> str:
        return f"<BaseGuild tag={self.tag!r} name={self.name!r}>"



from __future__ import annotations
from datetime import datetime
from typing import Union

from .base import CorkusBase
from .uuid import CorkusUUID

class MojangSkinResponse(CorkusBase):
    """Represents a response from Mojang API containing information about minecraft
    player skin. For more information see
    `Mojang documentation <https://wiki.vg/Mojang_API#UUID_to_Profile_and_Skin.2FCape>`_.
    """
    @property
    def requested(self) -> Union[datetime, None]:
        """Datetime when request to Mojang API actually was sent. Note that this property might be missing."""
        if self._attributes.get("timestamp") is not None:
            return datetime.fromtimestamp(self._attributes.get("timestamp", 0) / 1000)
        else:
            return None

    @property
    def uuid(self) -> Union[CorkusUUID, None]:
        """Minecraft UUID of player. Note that this property might be missing."""
        if self._attributes.get("profileId") is not None:
            return CorkusUUID(self._attributes.get("profileId"))
        else:
            return None

    @property
    def username(self) -> Union[str, None]:
        """Minecraft username of player. Note that this property might be missing."""
        return self._attributes.get("profileName")

    @property
    def url(self) -> str:
        """URL to the player's skin."""
        return self._attributes.get("textures", {}).get("SKIN", {}).get("url", "")

    def __repr__(self) -> str:
        return f"<MojangSkinResponse username={self.username!r} skin_url={self.url!r}>"



from __future__ import annotations
from typing import TYPE_CHECKING, Optional
from .base_guild import BaseGuild

if TYPE_CHECKING:
    from .guild import Guild

class LeaderboardGuild(BaseGuild):
    """Diffrent version of :py:class:`Guild` returned by :py:class:`LeaderboardEndpoint <corkus.endpoints.LeaderboardEndpoint>`."""
    @property
    def members_count(self) -> int:
        """Count of guild members, to access full list use :py:func:`fetch`."""
        return self._attributes.get("membersCount", 0)

    @property
    def total_xp(self) -> int:
        """Total sum of all xp points collected by this guild."""
        return self._attributes.get("xp", 0)

    @property
    def position(self) -> int:
        """This guild possition in guilds leaderboard."""
        return self._attributes.get("num", 0)

    @property
    def war_count(self) -> int:
        """Number of wars the guild has done, including failed ones."""
        return self._attributes.get("warCount", 0)

    async def fetch(self, timeout: Optional[int] = None) -> Guild:
        """Fetch information about this guild from Guilds API.

        .. include:: ../note_api_call.rst

        :param timeout: Optionally override default timeout.
        """
        return await self._corkus.guild.get(self.name, timeout)

    def __repr__(self) -> str:
        return f"<Guild tag={self.tag!r} name={self.name!r}>"



from __future__ import annotations
from .base import CorkusBase

class WeaponDamage(CorkusBase):
    """Damage dealt by a weapon."""
    @property
    def neutral(self) -> str:
        """Amount of neutral damage this weapon deals. Format: ``Minimum-Maximum``."""
        return self._attributes.get("damage", "0-0")

    @property
    def fire(self) -> str:
        """Amount of fire damage this weapon deals. Format: ``Minimum-Maximum``."""
        return self._attributes.get("fireDamage", "0-0")

    @property
    def water(self) -> str:
        """Amount of water damage this weapon deals. Format: ``Minimum-Maximum``."""
        return self._attributes.get("waterDamage", "0-0")

    @property
    def air(self) -> str:
        """Amount of air damage this weapon deals. Format: ``Minimum-Maximum``."""
        return self._attributes.get("airDamage", "0-0")

    @property
    def thunder(self) -> str:
        """Amount of thunder damage this weapon deals. Format: ``Minimum-Maximum``."""
        return self._attributes.get("thunderDamage", "0-0")

    @property
    def earth(self) -> str:
        """Amount of earth damage this weapon deals. Format: ``Minimum-Maximum``."""
        return self._attributes.get("earthDamage", "0-0")

    def __repr__(self) -> str:
        return (
            f"<WeaponDamage neutral={self.neutral} fire={self.fire} water={self.water} "
            f"air={self.air} thunder={self.thunder} earth={self.earth}>"
        )



from __future__ import annotations
from corkus.objects.base import CorkusBase
from typing import TYPE_CHECKING, Optional


if TYPE_CHECKING:
    from .ingredient import Ingredient

class PartialIngredient(CorkusBase):
    """Represents a ``Partial`` version of :py:class:`Ingredient`."""
    @property
    def name(self) -> str:
        """The name of the ingredient."""
        return self._attributes

    async def fetch(self, timeout: Optional[int] = None) -> Ingredient:
        """Fetch full ingredient information from API.

        .. include:: ../note_api_call.rst

        :param timeout: Optionally override default timeout.
        """
        return await self._corkus.ingredient.get(self.name, timeout)

    def __repr__(self) -> str:
        return f"<PartialIngredient name={self.name!r}>"



from __future__ import annotations
from corkus.objects.identification_type import IdentificationType
from typing import List, Union
from enum import Enum
import base64
import json
import math

from .base import CorkusBase
from corkus.utils import CorkusEnum
from .enums import ItemType, ItemCategory
from .player_class import ClassType
from .color import Color
from .quest import Quest
from .weapon_damage import WeaponDamage
from .skill_points import SkillPoints
from .armour_defense import ArmourDefence
from .major_identification import MajorIdentification
from .mojang_skin_response import MojangSkinResponse
from .identification_values import IdentificationValues
from .identification import Identification

class ItemTier(CorkusEnum):
    """Rarity tier of the item."""

    MYTHIC = "MYTHIC"
    FABLED = "FABLED"
    LEGENDARY = "LEGENDARY"
    RARE = "RARE"
    UNIQUE = "UNIQUE"
    SET = "SET"
    NORMAL = "NORMAL"

class ArmourType(CorkusEnum):
    """Material from which armour is made, same as in vanilla."""

    DIAMOND = "DIAMOND"
    GOLDEN = "GOLDEN"
    CHAIN = "CHAIN"
    IRON = "IRON"
    LEATHER = "LEATHER"

class ItemRestrictions(Enum):
    """Restrictions applied to some items."""

    UNTRADABLE = "UNTRADABLE"
    """Items that cannot be traded or dropped."""

    QUEST_ITEM = "QUEST_ITEM"
    """A special kind of :py:attr:`UNTRADABLE` items awarded during
    or after completing a quest."""

class AttackSpeed(CorkusEnum):
    """Attack speed of a weapon."""

    SUPER_FAST = "SUPER_FAST"
    VERY_FAST = "VERY_FAST"
    FAST = "FAST"
    NORMAL = "NORMAL"
    SLOW = "SLOW"
    SUPER_SLOW = "SUPER_SLOW"
    VERY_SLOW = "VERY_SLOW"

class Item(CorkusBase):
    """Represents regular (non-crafted) Wynncraft item."""
    @property
    def name(self) -> str:
        """The name of the item."""
        return self._attributes.get("name", "")

    @property
    def display_name(self) -> str:
        """Name that should be displayed to end-user, usually match :py:attr:`name`.

        .. caution::
            This property may not be reliable. See:
            `Wynncraft/WynncraftAPI19 <https://github.com/Wynncraft/WynncraftAPI/issues/19>`_.
        """
        return self._attributes.get("displayName", self.name)

    @property
    def type(self) -> ItemType:
        """Type of the item."""
        return ItemType(self._attributes.get("type", self._attributes.get("accessoryType", "Wand")).upper())

    @property
    def category(self) -> ItemCategory:
        """Category of the item."""
        return ItemCategory.from_type(self.type)

    @property
    def tier(self) -> ItemTier:
        """Item rarity tier."""
        return ItemTier(self._attributes.get("tier", "Normal").upper())

    @property
    def set(self) -> Union[str, None]:
        """The name of the set this item is part of. ``None`` if not part of a set.

        .. caution::
                This property is currently bugged and works only for ``Leaf`` set. See:
                `Wynncraft/WynncraftAPI#36 <https://github.com/Wynncraft/WynncraftAPI/issues/36>`_."""
        return self._attributes.get("set", None)

    @property
    def sockets(self) -> int:
        """Number of powder slots this item has."""
        return self._attributes.get("sockets", 0)

    @property
    def armour_type(self) -> Union[ArmourType, None]:
        """Material from which armour is made. If not a armour piece, or
        item is player head, returns ``None``."""
        if self.skin is not None:
            return None
        elif self._attributes.get("armorType", None) is None:
            return None
        else:
            return ArmourType(self._attributes.get("armorType").upper())

    @property
    def armour_color(self) -> Union[Color, None]:
        """Color that this armour piece is dyed. This is only present
        if :py:attr:`armour_type` is :py:attr:`ArmourType.LEATHER` or
        else returns ``None``."""
        if self.armour_type == ArmourType.LEATHER:
            return Color(tuple(int(c) for c in self._attributes.get("armorColor", "160,101,64").split(",")))
        else:
            return None

    @property
    def required_level(self) -> int:
        """Minimum combat level
        (:py:attr:`PlayerProfession.level <corkus.objects.PlayerProfession.level>` from
        :py:attr:`PlayerClass.combat <corkus.objects.PlayerClass.combat>`) required to use this item."""
        return self._attributes.get("level", 1)

    @property
    def required_class(self) -> Union[ClassType, None]:
        """The class that is required for this item to be used.
        ``None`` if item can be used by all classes."""
        if self._attributes.get("classRequirement") is not None:
            return ClassType(self._attributes.get("classRequirement").upper())
        elif self.type == ItemType.BOW:
            return ClassType.ARCHER
        elif self.type == ItemType.SPEAR:
            return ClassType.WARRIOR
        elif self.type == ItemType.WAND:
            return ClassType.MAGE
        elif self.type == ItemType.DAGGER:
            return ClassType.ASSASSIN
        elif self.type == ItemType.RELIK:
            return ClassType.SKYSEER
        else:
            return None

    @property
    def skill_points(self) -> SkillPoints:
        """Skill points required in order to use this item."""
        return SkillPoints(
            corkus = self._corkus,
            strength = self._attributes.get("strength", 0),
            dexterity = self._attributes.get("dexterity", 0),
            intelligence = self._attributes.get("intelligence", 0),
            defence = self._attributes.get("defence", 0),
            agility = self._attributes.get("agility", 0)
        )

    @property
    def required_quest(self) -> Union[Quest, None]:
        """The quest that must have been completed in order
        for this item to be used."""
        return self._attributes.get("quest", None)

    @property
    def restrictions(self) -> Union[ItemRestrictions, None]:
        """Restrictions applied to this item."""
        if self._attributes.get("restrictions", None) is None:
            return None
        else:
            return ItemRestrictions(self._attributes.get("restrictions").upper().replace(" ", "_"))

    @property
    def lore(self) -> Union[str, None]:
        """This item's lore."""
        return self._attributes.get("addedLore", None)

    @property
    def damage(self) -> Union[WeaponDamage, None]:
        """If this item is a weapon (:py:attr:`category` is :py:attr:`ItemCategory.WEAPON`)
        return it's damage."""
        if self.category is ItemCategory.WEAPON:
            return WeaponDamage(self._corkus, self._attributes)
        else:
            return None

    @property
    def attack_speed(self) -> Union[AttackSpeed, None]:
        """If this item is a weapon (:py:attr:`category` is :py:attr:`ItemCategory.WEAPON`)
        return it's attack speed."""
        if self.category is ItemCategory.WEAPON:
            return AttackSpeed(self._attributes.get("attackSpeed", "VERY_SLOW"))
        else:
            return None

    @property
    def identified(self) -> bool:
        """Is this item is pre-identified when obtained."""
        return self._attributes.get("identified", False)

    @property
    def health(self) -> Union[AttackSpeed, None]:
        """If this item is a armour piece (:py:attr:`category` is :py:attr:`ItemCategory.ARMOUR`)
        return set amount of health this armor provides."""
        if self.category is ItemCategory.ARMOUR:
            return self._attributes.get("health", 0)
        else:
            return None

    @property
    def armour_defence(self) -> Union[ArmourDefence, None]:
        """If this item is a armour piece (:py:attr:`category` is :py:attr:`ItemCategory.ARMOUR`)
        return protection it gives against elemental attacks."""
        if self.category is ItemCategory.ARMOUR:
            return ArmourDefence(self._corkus, self._attributes)
        else:
            return None

    @property
    def major_identifications(self) -> List[MajorIdentification]:
        """List all Major IDs on the item."""
        return [MajorIdentification(self._corkus, i) for i in self._attributes.get("majorIds", [])]

    @property
    def identifications(self) -> List[Identification]:
        """List all identifications of this item."""
        not_identified_id = (
            IdentificationType.AIR_DEFENSE,
            IdentificationType.FIRE_DEFENSE,
            IdentificationType.EARTH_DEFENSE,
            IdentificationType.WATER_DEFENSE,
            IdentificationType.THUNDER_DEFENSE
        )

        result = []
        for key, value in self._attributes.items():
            if not isinstance(value, int) or value == 0: continue
            type = IdentificationType.from_items_api(key)
            if type is None: continue
            if self.identified or type in not_identified_id:
                result.append(Identification(self._corkus, type, value = value))
            else:
                result.append(Identification(self._corkus, type, values = self._generate_identification_value(value)))
        return result

    def _generate_identification_value(self, base_value: int) -> IdentificationValues:
        """The values given by the API represent the 'base value' of that identification, when an item is
        actually identified a random number is generated for each non-zero identification and is then
        multiplied by that identification's base value. For identifications that have a positive base
        value, the random number can be between 0.3 and 1.3. For identifications that have a negative
        base value, the random number can be between 0.7 and 1.3. The result after multiplication is then
        rounded to the nearest integer, with a maximum value of -1 for negative identifications and a
        minimum value of 1 for positive identifications.
        """
        min_value, max_value = None, None
        if base_value > 0:
            min_value = max(self._normal_round(base_value * 0.3), 1)
            max_value = max(self._normal_round(base_value * 1.3), 1)
        else:
            min_value = min(self._normal_round(base_value * 0.7), -1)
            max_value = min(self._normal_round(base_value * 1.3), -1)
        return IdentificationValues(
            self._corkus,
            min = min_value,
            max = max_value
        )

    def _normal_round(self, n):
        if n - math.floor(n) < 0.5:
            return math.floor(n)
        return math.ceil(n)

    @property
    def skin(self) -> Union[MojangSkinResponse, None]:
        """If this item is a helmet (:py:attr:`type` is :py:attr:`ItemType.HELMET`)
        in form player head, this return information about owner of the head."""
        if self._attributes.get("skin") is not None:
            response = base64.b64decode(self._attributes.get("skin"))
            attributes = json.loads(response)
            return MojangSkinResponse(self._corkus, attributes)
        else:
            return None

    @property
    def item_id(self) -> str:
        """Minecraft `block`_/`item`_ ID + optional `data value`_,
        pre-flattening. Format: ``ID:DV``

        .. _block: https://minecraft.fandom.com/wiki/Java_Edition_data_values/Pre-flattening/Block_IDs
        .. _item: https://minecraft.fandom.com/wiki/Java_Edition_data_values/Pre-flattening/Item_IDs
        .. _data value: https://minecraft.fandom.com/wiki/Java_Edition_data_values/Pre-flattening"""
        if self.skin is not None:
            return "397:3"

        return self._attributes.get("material") or self._generate_id()

    def _generate_id(self) -> str:
        if self.armour_type == ArmourType.LEATHER:
            if self.type == ItemType.HELMET:
                return "298:0"
            elif self.type == ItemType.CHESTPLATE:
                return "299:0"
            elif self.type == ItemType.LEGGINGS:
                return "300:0"
            elif self.type == ItemType.BOOTS:
                return "301:0"
        elif self.armour_type == ArmourType.CHAIN:
            if self.type == ItemType.HELMET:
                return "302:0"
            elif self.type == ItemType.CHESTPLATE:
                return "303:0"
            elif self.type == ItemType.LEGGINGS:
                return "304:0"
            elif self.type == ItemType.BOOTS:
                return "305:0"
        elif self.armour_type == ArmourType.IRON:
            if self.type == ItemType.HELMET:
                return "306:0"
            elif self.type == ItemType.CHESTPLATE:
                return "307:0"
            elif self.type == ItemType.LEGGINGS:
                return "308:0"
            elif self.type == ItemType.BOOTS:
                return "309:0"
        elif self.armour_type == ArmourType.DIAMOND:
            if self.type == ItemType.HELMET:
                return "310:0"
            elif self.type == ItemType.CHESTPLATE:
                return "311:0"
            elif self.type == ItemType.LEGGINGS:
                return "312:0"
            elif self.type == ItemType.BOOTS:
                return "313:0"
        elif self.armour_type == ArmourType.GOLDEN:
            if self.type == ItemType.HELMET:
                return "314:0"
            elif self.type == ItemType.CHESTPLATE:
                return "315:0"
            elif self.type == ItemType.LEGGINGS:
                return "316:0"
            elif self.type == ItemType.BOOTS:
                return "317:0"
        return "0:0"

    def __repr__(self) -> str:
        return f"<Item name={self.name!r} type={self.type} tier={self.tier}>"



from __future__ import annotations
from enum import Enum

from .base import CorkusBase
from .uuid import CorkusUUID
from corkus.utils import CorkusEnum

class PlayerRank(Enum):
    """Player special rank."""
    ADMINISTRATOR = "ADMINISTRATOR"
    """Admins are either the owners or developers of Wynncraft."""

    MODERATOR = "MODERATOR"
    """Moderators enforce the Wynncraft rules on the forums, game servers, and Discord.
    They are also here to help players when needed."""

    BUILDER = "BUILDER"
    """The Builder tag is given to the builders who built the physical map."""

    ITEM = "ITEM"
    """Members of the Item Team create new items as well as balance existing ones."""

    GAME_MASTER = "GAME_MASTER"
    """The GM tag is given to Game Masters, the creators of items, quests, mobs, etc."""

    CMD = "CMD"
    """The CMD tag is given to the talented command blockers who create the cutscenes and
    puzzle you see in-game."""

    MUSIC = "MUSIC"
    """Given to the official composers of Wynncraft's music."""

    HYBRID = "HYBRID"
    """The Hybrid tag is given to Wynncraft Hybrids; Content Team members who are both builders, GMs, or CMDs."""

    MEDIA = "MEDIA"
    """Media is granted to those who record and post Wynncraft videos on Youtube or stream Wynncraft."""

    REGULAR = "REGULAR"

class PlayerTag(CorkusEnum):
    """Player's rank bought from `Wynncraft Store <https://store.wynncraft.com>`_."""
    CHAMPION = "CHAMPION"
    HERO = "HERO"
    VIP_PLUS = "VIP+"
    VIP = "VIP"
    PLAYER = "PLAYER"

class BasePlayer(CorkusBase):
    @property
    def uuid(self) -> CorkusUUID:
        """Minecraft UUID of player."""
        return CorkusUUID(self._attributes.get("uuid", ""))

    @property
    def rank(self) -> PlayerRank:
        """Player special rank, if don't have one, default to :py:attr:`PlayerRank.REGULAR`."""
        rank = self._attributes.get("rank")
        if rank is None or rank.upper() == "PLAYER":
            rank = "REGULAR"
        return PlayerRank(rank.upper().replace(" ", "_"))

    @property
    def tag(self) -> PlayerTag:
        """Player's rank bought from `Wynncraft Store <https://store.wynncraft.com>`_."""
        return PlayerTag(self._attributes.get("meta", {}).get("tag", {}).get("value", PlayerTag.PLAYER))

    @property
    def veteran(self) -> bool:
        """Is player a veteran, meaning, had VIP before August 1, 2014 which was when the 1.12 update was
        released and Minecraft EULA changed.
        """
        return self._attributes.get("meta", {}).get("veteran", False)

    def __repr__(self) -> str:
        return f"<BasePlayer uuid={self.uuid!r}>"



from __future__ import annotations
from typing import Union
from enum import Enum

class IdentificationType(Enum):
    # Additional skill points
    AGILITY_POINTS = "AGILITY_POINTS"
    """Additional agility points granted by this item."""
    DEFENSE_POINTS = "DEFENSE_POINTS"
    """Additional defence points granted by this item."""
    DEXTERITY_POINTS = "DEXTERITY_POINTS"
    """Additional dexterity points granted by this item."""
    INTELLIGENCE_POINTS = "INTELLIGENCE_POINTS"
    """Additional intelligence points granted by this item."""
    STRENGTH_POINTS = "STRENGTH_POINTS"
    """Additional strength points granted by this item."""

    # Attack
    ATTACK_SPEED = "ATTACK_SPEED"
    """Tier modifier to attack speed."""
    DAMAGE_BONUS = "DAMAGE_BONUS"
    """Percentage melee damage modifier."""
    DAMAGE_BONUS_RAW = "DAMAGE_BONUS_RAW"
    """Bonus to melee damage."""
    SPELL_DAMAGE = "SPELL_DAMAGE"
    """Percentage spell damage modifier."""
    SPELL_DAMAGE_RAW = "SPELL_DAMAGE_RAW"
    """Bonus to spell damage."""

    # Spells
    RAINBOW_SPELL_DAMAGE_RAW = "RAINBOW_SPELL_DAMAGE_RAW"
    """Bonus to rainbow spell damage."""
    SPELL_COST_1 = "SPELL_COST_1"
    """Modifier on cost of 1st spell."""
    SPELL_COST_1_RAW = "SPELL_COST_1_RAW"
    """Precentage modifier on cost of 1st spell."""
    SPELL_COST_2 = "SPELL_COST_2"
    """Modifier on cost of 2nd spell."""
    SPELL_COST_2_RAW = "SPELL_COST_2_RAW"
    """Precentage modifier on cost of 2nd spell."""
    SPELL_COST_3 = "SPELL_COST_3"
    """Modifier on cost of 3rd spell."""
    SPELL_COST_3_RAW = "SPELL_COST_3_RAW"
    """Precentage modifier on cost of 3rd spell."""
    SPELL_COST_4 = "SPELL_COST_4"
    """Modifier on cost of 4th spell."""
    SPELL_COST_4_RAW = "SPELL_COST_4_RAW"
    """Precentage modifier on cost of 4th spell."""

    # Elemental damage
    AIR_DAMAGE_BONUS = "AIR_DAMAGE_BONUS"
    """Percentage bonus to air damage."""
    EARTH_DAMAGE_BONUS = "EARTH_DAMAGE_BONUS"
    """Percentage bonus to earth damage."""
    FIRE_DAMAGE_BONUS = "FIRE_DAMAGE_BONUS"
    """Percentage bonus to fire damage."""
    THUNDER_DAMAGE_BONUS = "THUNDER_DAMAGE_BONUS"
    """Percentage bonus to thunder damage."""
    WATER_DAMAGE_BONUS = "WATER_DAMAGE_BONUS"
    """Percentage bonus to water damage."""

    # Elemental defence
    AIR_DEFENSE = "AIR_DEFENSE"
    """Air defense points granted by this item."""
    EARTH_DEFENSE = "EARTH_DEFENSE"
    """Earth defense points granted by this item."""
    FIRE_DEFENSE = "FIRE_DEFENSE"
    """Fire defense points granted by this item."""
    THUNDER_DEFENSE = "THUNDER_DEFENSE"
    """Thunder defense points granted by this item."""
    WATER_DEFENSE = "WATER_DEFENSE"
    """Wather defense points granted by this item."""

    AIR_DEFENSE_BONUS = "AIR_DEFENSE_BONUS"
    """Percentage bonus to air defense."""
    EARTH_DEFENSE_BONUS = "EARTH_DEFENSE_BONUS"
    """Percentage bonus to earth defense."""
    FIRE_DEFENSE_BONUS = "FIRE_DEFENSE_BONUS"
    """Percentage bonus to fire defense."""
    THUNDER_DEFENSE_BONUS = "THUNDER_DEFENSE_BONUS"
    """Percentage bonus to thunder defense."""
    WATER_DEFENSE_BONUS = "WATER_DEFENSE_BONUS"
    """Percentage bonus to wather defense."""

    # Health and mana
    HEALTH_BONUS = "HEALTH_BONUS"
    """Health modifier."""
    HEALTH_REGEN = "HEALTH_REGEN"
    """Percentage health regeneration modifier."""
    HEALTH_REGEN_RAW = "HEALTH_REGEN_RAW"
    """Bonus to health regeneration."""
    LIFE_STEAL = "LIFE_STEAL"
    """Amount of health stolen from enemy per attack across next 4 seconds."""
    MANA_STEAL = "MANA_STEAL"
    """Amount of mana stolen from enemy per attack across next 4 seconds."""
    MANA_REGEN = "MANA_REGEN"
    """Mana regeneration bonus per 4 seconds."""

    # Passive damage
    POISON = "POISON"
    """Percentage poison modifier."""
    REFLECTION = "REFLECTION"
    """Percentage reflection modifier."""
    THORNS = "THORNS"
    """Percentage thorns modifier."""
    EXPLODING = "EXPLODING"
    """Percentage exploding modifier."""

    # Movement
    WALK_SPEED = "WALK_SPEED"
    """Percentage walk speed modifier"""
    SPRINT = "SPRINT"
    """Percentage modifier on rate of depleting sprint bar."""
    SPRINT_REGEN = "SPRINT_REGEN"
    """Percentage sprint regeneration modifier"""
    JUMP_HEIGHT = "JUMP_HEIGHT"
    """Jump height modifier"""

    LOOT_BONUS = "LOOT_BONUS"
    """Percentage loot bonus modifier."""
    LOOT_QUALITY = "LOOT_QUALITY"
    """Percentage loot quality modifier."""
    XP_BONUS = "XP_BONUS"
    """Percentage XP bonus modifier."""
    EMERALD_STEALING = "EMERALD_STEALING"
    """Percentage bonus chance to steal an emerald on hit."""
    GATHER_XP_BONUS = "GATHER_XP_BONUS"
    """Percentage of additional XP when gartering."""
    GATHER_SPEED = "GATHER_SPEED"
    """Percentage modifier to gartering speed."""
    SOUL_POINTS = "SOUL_POINTS"
    """Percentage chance to regenerate an additional soul point."""

    @staticmethod
    def from_items_api(key: str) -> Union[IdentificationType, None]:
        from corkus.data import ids_convert
        return next((id["type"] for id in ids_convert if id["items_api"] == key), None)

    @staticmethod
    def from_ingredient_api(key: str) -> Union[IdentificationType, None]:
        from corkus.data import ids_convert
        return next((id["type"] for id in ids_convert if id["ingredient_api"] == key), None)

    @staticmethod
    def to_items_api(type: IdentificationType) -> str:
        from corkus.data import ids_convert
        return next((id["items_api"] for id in ids_convert if id["type"] == type), None)

    @staticmethod
    def to_ingredient_api(type: IdentificationType) -> str:
        from corkus.data import ids_convert
        return next((id["ingredient_api"] for id in ids_convert if id["type"] == type), None)



from __future__ import annotations
from typing import List

from .base import CorkusBase
from .partial_player import PartialPlayer
from .partial_guild import PartialGuild

class SearchResult(CorkusBase):
    """Combined result of search for :py:class:`Player` and :py:class:`Guild`."""
    @property
    def players(self) -> List[PartialPlayer]:
        """List of found players."""
        return [PartialPlayer(self._corkus, username = p) for p in self._attributes.get("players", [])]

    @property
    def guilds(self) -> List[PartialGuild]:
        """List of found guilds."""
        return [PartialGuild(self._corkus, g) for g in self._attributes.get("guilds", [])]

    @property
    def term(self) -> str:
        """Search term used to display these results."""
        return self._attributes.get("search", "")

    def __repr__(self) -> str:
        return f"<SearchResult players={self.players} guild={self.guilds}>"



from __future__ import annotations
from typing import List, Union

from .base import CorkusBase
from .server import Server
from .partial_online_player import PartialOnlinePlayer
from .player import Player
from .partial_player import PartialPlayer

class OnlinePlayers(CorkusBase):
    """List all running servers and players that are online on them."""
    @property
    def servers(self) -> List[Server]:
        """List all running servers."""
        return [Server(self._corkus, name, players) for name, players in self._attributes.items()]

    @property
    def players(self) -> List[PartialOnlinePlayer]:
        """List all online players."""
        return [p for s in self.servers for p in s.players]

    def get_player_server(self, player: Union[str, Player, PartialPlayer, PartialOnlinePlayer]) -> Union[Server, None]:
        """Get the server that player is currently playing on. Returns ``None`` if player is offline.

        :param player: Player username or player object to check for.
        """
        if isinstance(player, str):
            username = player
        elif isinstance(player, (Player, PartialOnlinePlayer)):
            username = player.username
        elif isinstance(player, PartialPlayer):
            username = player.username
            if username is None:
                raise ValueError("expected PartialPlayer with username")
        else:
            raise TypeError(f"expected str, Player, PartialPlayer or PartialOnlinePlayer, received {type(player)}")
        for p in self.players:
            if p.username == username:
                return p.server
        return None

    def is_player_online(self, player: Union[str, Player, PartialPlayer, PartialOnlinePlayer]) -> bool:
        """Check if player is online.

        :param player: Player username or player object to check for.
        """
        return self.get_player_server(player) is not None

    def __repr__(self) -> str:
        return f"<OnlinePlayers servers={len(self.servers)} players={len(self.players)}>"



# type: ignore

from .player import Player
from .partial_player import PartialPlayer
from .base_player import PlayerRank, PlayerTag
from .playtime import PlayerPlaytime, PlaytimeConvertRatio
from .player_statistics import PlayerStatistics, ClassStatistics
from .player_status import PlayerStatus
from .player_class import PlayerClass, ClassType
from .player_gamemodes import PlayerGamemodes, HardcoreType
from .leaderboard_player import LeaderboardPlayer
from .player_profession import PlayerProfession
from .dungeon import Dungeon, DungeonType
from .raid import Raid
from .quest import Quest
from .player_ranking import Ranking, PlayerRanking, PlayerOverallRanking, PlayerSoloRanking

from .guild import Guild
from .partial_guild import PartialGuild
from .leaderboard_guild import LeaderboardGuild
from .leaderboard_partial_member import LeaderboardPartialMember
from .guild_banner import GuildBanner, GuildBannerLayer, BannerColor, BannerPattern
from .member import Member, GuildRank
from .partial_member import PartialMember

from .territory import Territory
from .territory_location import TerritoryLocation
from .partial_teritories import PartialTeritories

from .server import Server, ServerType
from .partial_server import PartialServer
from .partial_online_player import PartialOnlinePlayer
from .online_players import OnlinePlayers

from .recipe import Recipe
from .partial_recipe import PartialRecipe

from .partial_ingredient import PartialIngredient
from .ingredient import Ingredient
from .material import Material
from .ingredient_sprite import IngredientSprite
from .ingredient_position import IngredientPositionModifiers
from .ingredient_comsumable import IngredientComsumableModifiers
from .ingredient_item import IngredientItemModifiers
from .level_range import LevelRange

from .item import Item, ArmourType, ItemTier, ItemRestrictions, AttackSpeed
from .mojang_skin_response import MojangSkinResponse
from .weapon_damage import WeaponDamage
from .armour_defense import ArmourDefence
from .major_identification import MajorIdentification

from .uuid import CorkusUUID
from .search_result import SearchResult
from .identification import Identification
from .identification_values import IdentificationValues
from .identification_type import IdentificationType
from .enums import ProfessionType, Timeframe, ItemType, LogicSymbol, ItemCategory
from .color import Color
from .skill_points import SkillPoints



from enum import Enum

class CorkusEnum(Enum):
    def __int__(self) -> int:
        for i, elem in enumerate(reversed(self.__class__)):
            if elem == self:
                return i
        return 0

    def __lt__(self, obj: object) -> bool:
        if not isinstance(obj, self.__class__):
            return NotImplemented
        return int(self) < int(obj)



from __future__ import annotations
from .base import CorkusBase

class IngredientComsumableModifiers(CorkusBase):
    """Modifier of :py:class:`Ingredient` applied to comsumables."""
    @property
    def duration(self) -> int:
        """Modifier on item duration."""
        return self._attributes.get("duration", 0)

    @property
    def charges(self) -> int:
        """Modifier on item charges."""
        return self._attributes.get("charges", 0)

    def __repr__(self) -> str:
        return f"<IngredientComsumableModifiers duration={self.duration} charges={self.charges}>"



from __future__ import annotations
from corkus.objects.base import CorkusBase
from typing import TYPE_CHECKING, Optional

if TYPE_CHECKING:
    from .guild import Guild

class PartialGuild(CorkusBase):
    @property
    def name(self) -> str:
        """The name of the guild."""
        return self._attributes

    async def fetch(self, timeout: Optional[int] = None) -> Guild:
        """Fetch full guild information from API.

        .. include:: ../note_api_call.rst

        :param timeout: Optionally override default timeout.
        """
        return await self._corkus.guild.get(self.name, timeout)

    def __repr__(self) -> str:
        return f"<PartialGuild name={self.name!r}>"



from __future__ import annotations
from corkus.objects.playtime import PlayerPlaytime
from typing import TYPE_CHECKING, Union, Optional

from .base_player import BasePlayer
from .leaderboard_partial_member import LeaderboardPartialMember
from .partial_guild import PartialGuild

if TYPE_CHECKING:
    from .player import Player

class LeaderboardPlayer(BasePlayer):
    """Diffrent version of :py:class:`Player` returned by :py:class:`LeaderboardEndpoint <corkus.endpoints.LeaderboardEndpoint>`."""
    @property
    def username(self) -> str:
        """Minecraft username of player."""
        return self._attributes.get("name", "")

    @property
    def playtime(self) -> PlayerPlaytime:
        """Time that player spent on wynncraft servers."""
        return PlayerPlaytime(self._attributes.get("minPlayed", 0))

    @property
    def pvp_kills(self) -> int:
        """Total player killed in `The Nether <https://wynncraft.fandom.com/wiki/The_Nether>`_."""
        return self._attributes.get("kills", 0)

    @property
    def total_combat_level(self) -> int:
        """Combined combat level across all classes."""
        return self._attributes.get("level", 0)

    @property
    def total_xp(self) -> int:
        """Total collected combat XP points across all classes."""
        return self._attributes.get("level", 0)

    @property
    def position(self) -> int:
        """This player possition in players leaderboard."""
        return self._attributes.get("num", 0)

    @property
    def member(self) -> Union[LeaderboardPartialMember, None]:
        """Partial representation of player in the guild."""
        if self._attributes.get("guildTag") is None:
            return None
        else:
            return LeaderboardPartialMember(
                corkus = self._corkus,
                uuid = self.uuid,
                username = self.username,
                guild = self.guild,
                tag = self._attributes.get("guildTag")
            )

    @property
    def guild(self) -> Union[PartialGuild, None]:
        """Partial information about player's guild."""
        if self._attributes.get("guild") is None:
            return None
        else:
            return PartialGuild(self._corkus, self._attributes.get("guild"))

    async def fetch(self, timeout: Optional[int] = None) -> Player:
        """Fetch information about this player from Players API.

        .. include:: ../note_api_call.rst

        :param timeout: Optionally override default timeout.
        """
        return await self._corkus.player.get(self.uuid, timeout)

    def __repr__(self) -> str:
        return f"<LeaderboardPlayer username={self.username!r}>"



from __future__ import annotations
from enum import Enum
from typing import Union

class PlaytimeConvertRatio(Enum):
    """Most popular playtime convertion ratios."""

    RAW = 1.0
    """1.0 value returned by the API."""

    OFFICIAL = 4.7
    """4.7 ratio used by official wynncraft statistics page."""

    WYNNDATA = 5
    """5.0 ratio used by `Wynndata <https://www.wynndata.tk/>`_."""

class PlayerPlaytime():
    """Object that helps to convert playtime

    Wynncraft API returns player playtime in wired way and it can't be used directly.
    See `Wynncraft/WynncraftAPI#8 <https://github.com/Wynncraft/WynncraftAPI/issues/8>`_

    To get most popular conversion ratios see :py:class:`PlaytimeConvertRatio` Enum.
    """
    def __init__(self, playtime: int) -> None:
        self._playtime = playtime

    @property
    def raw(self) -> int:
        """Get number that API returned."""
        return self._playtime

    def seconds(self, ratio: Union[PlaytimeConvertRatio, float]) -> int:
        """Convert time using specified ratio and return it as seconds.

        :param ratio: Ratio by which raw playtime will be multiplied.
        """
        return self._convert(ratio) * 60

    def minutes(self, ratio: Union[PlaytimeConvertRatio, float]) -> int:
        """Convert time using specified ratio and return it as minutes.

        :param ratio: Ratio by which raw playtime will be multiplied.
        """
        return self._convert(ratio)

    def hours(self, ratio: Union[PlaytimeConvertRatio, float]) -> int:
        """Convert time using specified ratio and return it as hours.

        :param ratio: Ratio by which raw playtime will be multiplied.
        """
        return self._convert(ratio) / 60

    def _convert(self, ratio: Union[PlaytimeConvertRatio, float]) -> int:
        if isinstance(ratio, PlaytimeConvertRatio):
            ratio = ratio.value
        return round(self._playtime * ratio)



from __future__ import annotations
from .base import CorkusBase

class Material(CorkusBase):
    """Represents crafting material used in crafting items using :py:class:`Recipe`."""

    @property
    def name(self) -> str:
        """Name of the material like ``Refined Copper Ingot`` or ``Refined Wheat String``."""
        return self._attributes.get("item", "")

    @property
    def amount(self) -> int:
        """Amount of this material to craft that item."""
        return self._attributes.get("amount", 0)

    def __repr__(self) -> str:
        return f"<Material name={self.name!r} amount={self.amount}>"



from __future__ import annotations
from typing import List, TYPE_CHECKING
from enum import Enum
from .base import CorkusBase
from .partial_online_player import PartialOnlinePlayer

if TYPE_CHECKING:
    from corkus import Corkus

class ServerType(Enum):
    """Type of :py:class:`Server`."""
    REGULAR = "WC"
    """Regular Wynncraft servers."""

    YOUTUBE = "YT"
    """Servers restricted to players with :py:attr:`PlayerRank.MEDIA` rank."""

class Server(CorkusBase):
    """Represents a singular Wynncraft Server."""
    def __init__(self, corkus: Corkus, name: str, players: List[str]):
        self._name = name
        super().__init__(corkus, players)

    @property
    def name(self) -> str:
        """The name of server like ``WC1`` or ``WC16`` or ``YT``."""
        return self._name

    @property
    def players(self) -> List[PartialOnlinePlayer]:
        """List of all online players on this server."""
        return [PartialOnlinePlayer(self._corkus, self, username = p) for p in self._attributes]

    @property
    def total_players(self) -> int:
        """Total number of online player on this server."""
        return len(self.players)

    @property
    def type(self) -> ServerType:
        """Type of the server, most servers are stanard servers."""
        return ServerType("".join([i for i in self._name if not i.isdigit()]))

    @property
    def regular(self) -> bool:
        """Is this server a regular server (:py:attr:`type` is :py:attr:`ServerType.REGULAR`)."""
        return self.type == ServerType.REGULAR

    def __repr__(self) -> str:
        return f"<Server name={self.name!r} total_players={self.total_players} players={self.players}>"



from __future__ import annotations
from typing import List, Optional

from .endpoint import Endpoint
from corkus.utils.request import APIVersion
from corkus.objects import PartialGuild, Guild

class GuildEndpoint(Endpoint):
    async def list_all(self, timeout: Optional[int] = None) -> List[PartialGuild]:
        """List all active guild on the server.

        :param timeout: Optionally override default timeout.
        """
        response = await self._request.get(
            version = APIVersion.V1,
            parameters = "guildList",
            timeout = timeout
        )
        return [PartialGuild(self._corkus, g) for g in response.get("guilds", [])]

    async def get(self, name: str, timeout: Optional[int] = None) -> Guild:
        """Get statics of the guild by given name.

        :param timeout: Optionally override default timeout.
        """
        response = await self._request.get(
            version = APIVersion.V1,
            parameters = "guildStats&command=" + name,
            timeout = timeout
        )
        return Guild(self._corkus, response)

    async def search(self, term: str, timeout: Optional[int] = None) -> List[PartialGuild]:
        """Search for guilds using specified search term.

        :param term: Search term for guild search.
        :param timeout: Optionally override default timeout.
        """
        result = await self._corkus.search.guilds_and_players(term, timeout)
        return result.guilds



from __future__ import annotations
from typing import Optional, TYPE_CHECKING
from corkus.utils.request import CorkusRequest

from corkus.endpoints import (
    NetworkEndpoint,
    PlayerEndpoint,
    GuildEndpoint,
    TerritoryEndpoint,
    SearchEndpoint,
    IngredientEndpoint,
    LeaderboardEndpoint,
    RecipeEndpoint,
    ItemEndpoint
)

if TYPE_CHECKING:
    from aiohttp import ClientSession

class RateLimit:
    """Current ratelimit information for a Corkus instance. You should considier
    these values as a information rather than use them in regulating your requests
    since, Corkus have a ratelimit system in place.
    """
    def __init__(self, total: int, remaining: int, reset: int) -> None:
        self._total = total
        self._remaining = remaining
        self._reset = reset

    @property
    def total(self) -> int:
        """Total number of requests that user can send per minute under provided
        API key or IP if one is not provided."""
        return self._total

    @property
    def remaining(self) -> int:
        """Total number of requests left that user can send for that minute
        (:py:attr:`total` - used)."""
        return self._remaining

    @property
    def reset(self) -> int:
        """Number of seconds left until this minute ends and thus ratelimit
        is restored back to :py:attr:`total`."""
        return self._reset

class Corkus:
    """First-class interface for accessing Wynncraft API.

    For configuration options see: :ref:`configuration`."""

    def __init__(self, *,
        api_key: Optional[str] = None,
        timeout: Optional[int] = 30,
        session: Optional[ClientSession] = None,
        ratelimit_enable: Optional[bool] = True,
        cache_enable: Optional[bool] = True,
    ) -> None:
        self._api_key = api_key
        self._request = CorkusRequest(timeout, api_key, session, ratelimit_enable, cache_enable)

    async def __aenter__(self) -> "Corkus":
        """Async enter."""
        return self

    async def __aexit__(self, *exc_info) -> None:
        """Async exit."""
        await self.close()

    @property
    def player(self) -> PlayerEndpoint:
        """General statistics of wynncraft players."""
        return PlayerEndpoint(self, self._request)

    @property
    def guild(self) -> GuildEndpoint:
        """Information about server guilds."""
        return GuildEndpoint(self, self._request)

    @property
    def network(self) -> NetworkEndpoint:
        """Wynncraft Network specific routes like list of all players."""
        return NetworkEndpoint(self, self._request)

    @property
    def territory(self) -> TerritoryEndpoint:
        """Information about teritories."""
        return TerritoryEndpoint(self, self._request)

    @property
    def search(self) -> SearchEndpoint:
        """Search for guild and players."""
        return SearchEndpoint(self, self._request)

    @property
    def item(self) -> ItemEndpoint:
        """Regular (not crafted) items database."""
        return ItemEndpoint(self, self._request)

    @property
    def recipe(self) -> RecipeEndpoint:
        """Crafted items statistics and recipes."""
        return RecipeEndpoint(self, self._request)

    @property
    def ingredient(self) -> IngredientEndpoint:
        """Information about ingredients."""
        return IngredientEndpoint(self, self._request)

    @property
    def leaderboard(self) -> LeaderboardEndpoint:
        """Leaderboards of best players and guilds."""
        return LeaderboardEndpoint(self, self._request)

    @property
    def rate_limit(self) -> RateLimit:
        """Current ratelimit information for this Corkus instance."""
        return RateLimit(
            total = self._request.ratelimit.total,
            remaining = self._request.ratelimit.remaining,
            reset = self._request.ratelimit.reset
        )

    async def close(self) -> None:
        """End the corkus client when it's not needed anymore."""
        return await self._request.close()



from __future__ import annotations
from .base import CorkusBase

class ClassStatistics(CorkusBase):
    """General Statistics of a :py:class:`PlayerClass`"""
    @property
    def chests_found(self) -> int:
        """Total chest opened."""
        return self._attributes.get("chestsFound", 0)

    @property
    def blocks_walked(self) -> int:
        """Total blocks walked."""
        return self._attributes.get("blocksWalked", 0)

    @property
    def items_identified(self) -> int:
        """Total items identified.

        .. caution::
                This property is currently bugged and always returns ``0``. See:
                `Wynncraft/WynncraftAPI#62 <https://github.com/Wynncraft/WynncraftAPI/issues/62>`_.
        """
        return self._attributes.get("itemsIdentified", 0)

    @property
    def mobs_killed(self) -> int:
        """Total mobs killed."""
        return self._attributes.get("mobsKilled", 0)

    @property
    def pvp_kills(self) -> int:
        """Total player killed in `The Nether <https://wynncraft.fandom.com/wiki/The_Nether>`_."""
        return self._attributes.get("pvp", {}).get("kills", 0)

    @property
    def pvp_deaths(self) -> int:
        """Total deaths by players in `The Nether <https://wynncraft.fandom.com/wiki/The_Nether>`_."""
        return self._attributes.get("pvp", {}).get("deaths", 0)

    @property
    def deaths(self) -> int:
        """Number of deaths not caused by PVP."""
        return self._attributes.get("deaths", 0)

    @property
    def discoveries(self) -> int:
        """Total number of discoveries by player."""
        return self._attributes.get("discoveries", 0)

    @property
    def swarms_won(self) -> int:
        """Total number of `Swarm Events <https://wynncraft.fandom.com/wiki/Swarm>`_ won by player."""
        return self._attributes.get("eventsWon", 0)

    @property
    def logins(self) -> int:
        """Number of logins to the server."""
        return self._attributes.get("deaths", 0)

    def __repr__(self) -> str:
        return f"<ClassStatistics chests_found={self.chests_found} mobs_killed={self.mobs_killed}>"

class PlayerStatistics(ClassStatistics):
    """General Statistics of a :py:class:`Player`. This class is based on :py:class:`ClassStatistics`."""
    @property
    def total_combat_level(self) -> int:
        """Sum of combat level across all classes."""
        return self._attributes.get("totalLevel", {}).get("combat", 0)

    @property
    def total_profession_level(self) -> int:
        """Sum of profession level across all classes."""
        return self._attributes.get("totalLevel", {}).get("profession", 0)

    def __repr__(self) -> str:
        return f"<PlayerStatistics chests_found={self.chests_found} mobs_killed={self.mobs_killed}>"



from __future__ import annotations
from typing import Literal, List

from .base import CorkusBase
from .enums import ProfessionType
from .ingredient_sprite import IngredientSprite
from .ingredient_position import IngredientPositionModifiers
from .ingredient_comsumable import IngredientComsumableModifiers
from .ingredient_item import IngredientItemModifiers
from .identification import Identification
from .identification_values import IdentificationValues
from .identification_type import IdentificationType

class Ingredient(CorkusBase):
    """Crafting Ingredients are items found in the world of
    Wynncraft used to add extra effects onto crafted items."""
    @property
    def name(self) -> str:
        """The name of the ingredient."""
        return self._attributes.get("name", "")

    @property
    def tier(self) -> Literal[0, 1, 2, 3]:
        """Number of stars on the ingredient."""
        return self._attributes.get("tier", 0)

    @property
    def required_level(self) -> int:
        """Level that an player must get on all of the :py:attr:`~required_professions` in order to use this ingredient."""
        return self._attributes.get("level", 0)

    @property
    def required_professions(self) -> List[ProfessionType]:
        """List of professions on which player must get :py:attr:`~required_level` in order to use this ingredient."""
        return [ProfessionType(p) for p in self._attributes.get("skills", {})]

    @property
    def sprite(self) -> IngredientSprite:
        """Describes look of ingredient in-game."""
        return IngredientSprite(self._corkus, self._attributes.get("sprite", {}))

    @property
    def identifications(self) -> List[Identification]:
        """List all identifications added by this ingredient."""
        result = []
        for key, value in self._attributes.get("identifications", []).items():
            type = IdentificationType.from_ingredient_api(key)
            if type is None:
                raise ValueError(f"unknown identification: {key}")
            result.append(Identification(self._corkus, type,
                values = IdentificationValues(self._corkus, attributes = value)
            ))
        return result

    @property
    def position_modifiers(self) -> IngredientPositionModifiers:
        """How this ingredient affect other ones in the grid."""
        return IngredientPositionModifiers(self._corkus, self._attributes.get("ingredientPositionModifiers", {}))

    @property
    def item_modifiers(self) -> IngredientItemModifiers:
        """Additional modifiers applied only on not consumable items like armour, weapons or jewelry."""
        return IngredientItemModifiers(self._corkus, self._attributes.get("itemOnlyIDs", {}))

    @property
    def consumable_modifiers(self) -> IngredientComsumableModifiers:
        """Additional modifiers applied only on comsumable items like food, potions and scrolls."""
        return IngredientComsumableModifiers(self._corkus, self._attributes.get("consumableOnlyIDs", {}))

    def __repr__(self) -> str:
        return f"<Ingredient name={self.name!r}>"



from __future__ import annotations
from typing import TYPE_CHECKING, List, Optional, Union

from .endpoint import Endpoint
from corkus.utils.request import APIVersion
from corkus.objects import Player, CorkusUUID

if TYPE_CHECKING:
    from corkus.objects import PartialPlayer

class PlayerEndpoint(Endpoint):
    async def get(self, username_or_uuid: Union[str, CorkusUUID], timeout: Optional[int] = None) -> Player:
        """Get statistics of specific player.

        .. note::
            If you are getting player by UUID please make sure it's in
            the dashed form or, even better, wrap it intro :py:class:`CorkusUUID <corkus.objects.CorkusUUID>`
            like so:

            .. code-block:: python

                corkus.player.get(CorkusUUID(uuid))

        :param username_or_uuid: Username or UUID of the player.
        :param timeout: Optionally override default timeout.
        """
        if isinstance(username_or_uuid, CorkusUUID):
            username_or_uuid = username_or_uuid.string(dashed = True)

        response = await self._request.get(
            version = APIVersion.V2,
            parameters = f"player/{username_or_uuid}/stats",
            timeout = timeout
        )
        return Player(self._corkus, response.get("data", {})[0])

    async def get_uuid(self, username: str, timeout: Optional[int] = None) -> CorkusUUID:
        """Get UUID from player username

        .. danger::
            This should not be used in place of the Mojang API, and your IP will be blocked if you do so.

        :param username: Username of the player.
        :param timeout: Optionally override default timeout.
        """
        response = await self._request.get(
            version = APIVersion.V2,
            parameters = f"player/{username}/uuid",
            timeout = timeout
        )
        return CorkusUUID(response.get("data", [])[0].get("uuid", ""))

    async def search(self, term: str, timeout: Optional[int] = None) -> List[PartialPlayer]:
        """Search for players using specified search term.

        :param username: Search term.
        :param timeout: Optionally override default timeout.
        """
        result = await self._corkus.search.guilds_and_players(term, timeout)
        return result.players



from __future__ import annotations
from typing import Dict, Union
from .base import CorkusBase
from .enums import ProfessionType

class Ranking(CorkusBase):
    """Represents ranking information of :py:class:`Player`."""
    @property
    def player(self) -> PlayerRanking:
        """Player specific ranking."""
        return PlayerRanking(self._corkus, self._attributes.get("player", {}))

    @property
    def pvp(self) -> None:
        """Pvp ranking of the player.

        .. warning::
            This functionality is not yet implemented intro Wynncraft API and will
            always return ``None``."""
        return self._attributes.get("pvp")

    @property
    def guild(self) -> None:
        """Ranking of player's guild.

        .. warning::
            This functionality is not yet implemented intro Wynncraft API and will
            always return ``None``.
        """
        return self._attributes.get("guild")

class PlayerRanking(CorkusBase):
    """Represents player-specific :py:class:`Ranking` information."""
    @property
    def solo(self) -> PlayerSoloRanking:
        """Solo ranking."""
        return PlayerSoloRanking(self._corkus, self._attributes.get("solo", {}))

    @property
    def overall(self) -> PlayerOverallRanking:
        """Overall ranking."""
        return PlayerOverallRanking(self._corkus, self._attributes.get("overall", {}))

class PlayerSoloRanking(CorkusBase):
    """Represents *solo* category of :py:class:`PlayerRanking`."""
    @property
    def combat(self) -> Union[int, None]:
        """Shortcut to combat in :py:attr:`professions`."""
        return self._attributes.get("combat")

    @property
    def professions(self) -> Dict[ProfessionType, Union[int, None]]:
        """Dictionary containing all of the :py:class`ProfessionType` as keys and
        player's places in these leaderboards as keys. Keys are ``None`` if not in top 100."""
        result = {}
        for prof in ProfessionType:
            result[prof] = self._attributes.get(prof.value.lower())
        return result

    @property
    def overall(self) -> Union[int, None]:
        """Player's place in overall solo leaderboard. Returns ``None`` if not in top 100."""
        return self._attributes.get("overall")

class PlayerOverallRanking(CorkusBase):
    """Represents *overall* category of :py:class:`PlayerRanking`."""
    @property
    def all(self) -> Dict[ProfessionType, Union[int, None]]:
        """Player's place in overall all leaderboard. Returns ``None`` if not in top 100."""
        return self._attributes.get("all")

    @property
    def professions(self) -> Union[int, None]:
        """Player's place in overall professions leaderboard. Returns ``None`` if not in top 100."""
        return self._attributes.get("profession")

    @property
    def combat(self) -> Union[int, None]:
        """Player's place in overall combat leaderboard. Returns ``None`` if not in top 100."""
        return self._attributes.get("combat")



from __future__ import annotations
from .territory import Territory
from typing import TYPE_CHECKING, Optional, List
from .partial_base import PartialBase

if TYPE_CHECKING:
    from corkus import Corkus
    from .guild import Guild

class PartialTeritories(PartialBase):
    """This ``Partial`` object is returned as a teritories of :py:class:`Guild`.
    API returns only numbers of teritories owned by a specific guild.
    You can use :py:func:`fetch` function which calls :py:func:`TerritoryEndpoint.list_all() <corkus.endpoints.TerritoryEndpoint.list_all>`
    and return detailed list of teritories owned by this guild."""

    def __init__(self, corkus: Corkus, guild: Guild, count: int):
        super().__init__(corkus)
        self._count = count
        self._guild = guild

    @property
    def count(self) -> int:
        """Total count of teritories owned by this guild."""
        return self._count

    def __len__(self) -> int:
        return self.count

    async def fetch(self, timeout: Optional[int] = None) -> List[Territory]:
        """Fetch full Territory information from API.

        .. include:: ../note_api_call.rst

        :param timeout: Optionally override default timeout.
        """
        teritories = await self._corkus.territory.list_all(timeout)
        return [t for t in teritories if t.guild.name == self._guild.name]



from __future__ import annotations
from typing import TYPE_CHECKING
from enum import Enum
from .base import CorkusBase

if TYPE_CHECKING:
    from corkus import Corkus

class HardcoreType(Enum):
    """Describes current status of player's hardcore challenge."""
    ENABLED = "ENABLED"
    DISABLED = "DISABLED"
    FAILED = "FAILED"

class PlayerGamemodes(CorkusBase):
    """Challenge gamemodes that are enabled on a :py:class:`PlayerClass`."""
    def __init__(self, corkus: Corkus, attributes: dict, deaths: int):
        super().__init__(corkus, attributes)
        self._deaths = deaths

    @property
    def craftsman(self) -> str:
        """Does this class have the craftsman challenge enabled."""
        return self._attributes.get("craftsman", False)

    @property
    def hardcore(self) -> HardcoreType:
        """Does this class have the hardcore challenge enabled and what is the status of it."""
        enabled = self._attributes.get("hardcore", False)
        active = self._deaths <= 0
        if enabled and active:
            return HardcoreType.ENABLED
        elif enabled and not active:
            return HardcoreType.FAILED
        else:
            return HardcoreType.DISABLED

    @property
    def ironman(self) -> bool:
        """Does this class have the ironman challenge enabled."""
        return self._attributes.get("ironman", False)

    @property
    def hunted(self) -> bool:
        """Does this class have the haunted mode enabled."""
        return self._attributes.get("hunted", False)

    def __repr__(self) -> str:
        return f"<Gamemodes craftsman={self.craftsman!r} hardcore={self.hardcore!r} ironman={self.ironman!r} hunted={self.hunted!r}>"



from __future__ import annotations
from typing import TYPE_CHECKING, Optional, Union

from .partial_base import PartialBase

if TYPE_CHECKING:
    from corkus import Corkus
    from .uuid import CorkusUUID
    from .player import Player

class PartialPlayer(PartialBase):
    """Represents a ``Partial`` version of :py:class:`Player`.
    Please note that :py:attr:`username` or :py:attr:`uuid` might
    be ``None`` but always at least one is present."""
    def __init__(self, corkus: Corkus, *, uuid: Optional[CorkusUUID] = None, username: Optional[str] = None):
        super().__init__(corkus)
        self._uuid = uuid
        self._username = username
        if uuid is None and username is None:
            raise ValueError("uuid and username are both none")

    @property
    def username(self) -> Union[str, None]:
        """Minecraft username of player."""
        return self._username

    @property
    def uuid(self) -> Union[CorkusUUID, None]:
        """Minecraft UUID of player."""
        return self._uuid

    async def fetch(self, timeout: Optional[int] = None) -> Player:
        """Fetch full player data from API.

        .. include:: ../note_api_call.rst

        :param timeout: Optionally override default timeout.
        """
        identifier = self._uuid if self._uuid is not None else self._username
        return await self._corkus.player.get(identifier, timeout)

    def __repr__(self) -> str:
        result = "<PartialPlayer"
        if self._uuid is not None:
            result += f" uuid={self._uuid.string()!r}"
        if self._username is not None:
            result += f" username={self._username!r}"
        return result + ">"



from __future__ import annotations
from .base import CorkusBase

class TerritoryLocation(CorkusBase):
    """The start and end coordinates of the :py:class:`Territory` area."""
    @property
    def start_x(self) -> int:
        return self._attributes.get("startX", 0)

    @property
    def start_y(self) -> int:
        return self._attributes.get("startY", 0)

    @property
    def end_x(self) -> int:
        return self._attributes.get("endX", 0)

    @property
    def end_y(self) -> int:
        return self._attributes.get("endY", 0)

    def __repr__(self) -> str:
        return f"<TerritoryLocation start_x={self.start_x} start_y={self.start_y} end_x={self.end_x} end_y={self.end_y}>"



from __future__ import annotations
from typing import TYPE_CHECKING
import iso8601
from datetime import datetime

from .base import CorkusBase
from corkus.utils import CorkusEnum
from .uuid import CorkusUUID
from .partial_player import PartialPlayer

if TYPE_CHECKING:
    from .guild import Guild
    from corkus import Corkus
    from .player import Player

class GuildRank(CorkusEnum):
    OWNER = "OWNER"
    CHIEF = "CHIEF"
    STRATEGIST = "STRATEGIST"
    CAPTAIN = "CAPTAIN"
    RECRUITER = "RECRUITER"
    RECRUIT = "RECRUIT"

class Member(CorkusBase):
    """Represents a member of a :py:class:`Guild`."""
    def __init__(self, corkus: Corkus, guild: Guild, attributes: dict):
        self._guild = guild
        super().__init__(corkus, attributes)

    @property
    def username(self) -> int:
        """Minecraft username of player."""
        return self._attributes.get("name", "")

    @property
    def uuid(self) -> CorkusUUID:
        """Minecraft UUID of player."""
        return CorkusUUID(self._attributes.get("uuid", ""))

    @property
    def contributed_xp(self) -> int:
        """Total number of contributed experience points by this player."""
        return self._attributes.get("contributed", 0)

    @property
    def join_date(self) -> datetime:
        """Datetime when player joined the guild."""
        return iso8601.parse_date(self._attributes.get("join", "1970"))

    @property
    def guild(self) -> Guild:
        """The guild that this member is a part of."""
        return self._guild

    @property
    def player(self) -> PartialPlayer:
        """Reduce to to :class:`PartialPlayer`."""
        return PartialPlayer(self._corkus, uuid = self.uuid, username = self.username)

    async def fetch_player(self) -> Player:
        """Fetch player data of this member from API.

        .. include:: ../note_api_call.rst"""
        return await self._corkus.player.get(self.uuid)



from __future__ import annotations
from .member import Member
from typing import List
from .base_guild import BaseGuild

class Guild(BaseGuild):
    """`Guilds <https://wynncraft.fandom.com/wiki/Guilds>`_ are Wynncraft communities of people
    that work together to achieve their goals.
    """
    @property
    def members(self) -> List[Member]:
        """List of all members currently in guild."""
        return [Member(self._corkus, self , m) for m in self._attributes.get("members", [])]

    @property
    def level_progress(self) -> float:
        """Progress to next :py:attr:`level` in precentage 0-100%

        .. caution::
            This property is currently bugged and return invalid precentage. See:
            `Wynncraft/WynncraftAPI#61 <https://github.com/Wynncraft/WynncraftAPI/issues/61>`_.
        """
        return self._attributes.get("xp", 0)

    def __repr__(self) -> str:
        return f"<Guild tag={self.tag!r} name={self.name!r}>"



from __future__ import annotations
from .base import CorkusBase

class IngredientSprite(CorkusBase):
    """Represents visual look of :py:class:`Ingredient`"""
    @property
    def id(self) -> int:
        """Minecraft item ID that represents this ingredient."""
        return self._attributes.get("id", 0)

    @property
    def damage(self) -> int:
        """I honestly don't have idea what this property does, if you know
        please `create a issue <https://github.com/MrBartusek/corkus.py/issues/new/choose>`_"""
        return self._attributes.get("damage", 0)

    def __repr__(self) -> str:
        return f"<IngredientSprite id={self.id} damage={self.damage}>"



from __future__ import annotations
from typing import Optional
from uuid import UUID

class CorkusUUID(UUID):
    """Simple class that overlays `UUID <https://docs.python.org/3/library/uuid.html>`_ and simplify it's conversion."""
    def __init__(self, uuid: str) -> None:
        super().__init__(uuid)

    def string(self, dashed: Optional[bool] = True) -> str:
        """Convert UUID to dashed or not dashed string

        :param dashed: Should UUID be dashed."""
        if dashed:
            return super().__str__()
        else:
            return str(super().hex)



from __future__ import annotations
from .base import CorkusBase

class Raid(CorkusBase):
    """Represents a `Raid <https://wynncraft.fandom.com/wiki/Raids>`_ completed by A :py:class:`Player`"""
    @property
    def name(self) -> str:
        """Name of the raid."""
        return self._attributes.get("name", "")

    @property
    def completed(self) -> int:
        """Total runs completed by the player. Failed runs are not counted."""
        return self._attributes.get("completed", 0)

    def __repr__(self) -> str:
        return f"<Raid name={self.name!r} completed={self.completed}>"



from __future__ import annotations
from typing import List, Optional

from .endpoint import Endpoint
from corkus.utils.request import APIVersion
from corkus.errors import InvalidInputError
from corkus.objects import Item, ItemType

class ItemEndpoint(Endpoint):
    async def get_all(self, timeout: Optional[int] = None) -> List[Item]:
        """List all available items.

        .. include:: ../note_big_endpoint.rst

        :param timeout: Optionally override default timeout.
        """
        response = await self._request.get(
            version = APIVersion.V1,
            parameters = "itemDB&category=all",
            timeout = timeout
        )
        return [Item(self._corkus, i) for i in response.get("items", [])]

    async def search_by_type(self, type: ItemType, timeout: Optional[int] = None) -> List[Item]:
        """Search for the items using their type
        (:py:attr:`Item.type <corkus.objects.Item.type>`).

        .. py:currentmodule:: corkus.objects

        .. warning::
            This function does not accept crafted items: :py:attr:`ItemType.SCROLL`,
            :py:attr:`ItemType.POTION` and :py:attr:`ItemType.FOOD`. Providing them
            will raise :py:exc:`corkus.errors.InvalidInputError`

        :param type: The type of returned items.
        :param timeout: Optionally override default timeout.
        :raises InvalidInputError: When invalid type is provided.
        """
        if type in (ItemType.SCROLL, ItemType.FOOD, ItemType.POTION):
            raise InvalidInputError(f"Searching for crafteds items, like {type}, is not possible.")

        response = await self._request.get(
            version = APIVersion.V1,
            parameters = "itemDB&category=" + type.value.lower(),
            timeout = timeout
        )
        return [Item(self._corkus, i) for i in response.get("items", [])]

    async def search_by_name(self, name: str, timeout: Optional[int] = None) -> List[Item]:
        """Search for the items using their name
        (:py:attr:`Item.name <corkus.objects.Item.name>`).

        :param name: Search term.
        :param timeout: Optionally override default timeout.
        """
        response = await self._request.get(
            version = APIVersion.V1,
            parameters = "itemDB&search=" + name,
            timeout = timeout
        )
        return [Item(self._corkus, i) for i in response.get("items", [])]



from __future__ import annotations
from typing import TYPE_CHECKING

from corkus.objects.base_partial_member import BasePartialMember

if TYPE_CHECKING:
    from corkus import Corkus
    from .uuid import CorkusUUID
    from .member import GuildRank
    from .partial_guild import PartialGuild

class PartialMember(BasePartialMember):
    """Represents a ``Partial`` version of :py:class:`Member`."""
    def __init__(self,
        corkus: Corkus,
        uuid: CorkusUUID,
        username: str,
        guild: PartialGuild,
        rank: GuildRank
        ):
        super().__init__(corkus, uuid, username, guild)
        self._rank = rank

    @property
    def rank(self) -> GuildRank:
        """Player's rank in guild."""
        return self._rank

    def __repr__(self) -> str:
        return f"<PartialMember username={self._username!r} guild={self._guild}>"



from __future__ import annotations
from .base import CorkusBase
from enum import Enum

class DungeonType(Enum):
    REMOVED = "REMOVED"
    """Dungeons that were removed from the game in version ``1.14.1`` like ``Skeleton`` or ``Spider``"""

    REMOVED_MINI = "REMOVED_MINI"
    """Minidungeons that were reworked in version ``1.17`` like ``Ice`` or ``Ocean``"""

    STANDARD = "STANDARD"
    """Generic dungeons like ``Galleon's Graveyard`` or ``Fallen Factory``"""

    CORRUPTED = "CORRUPTED"
    """Harder variant of standard dungeons like ``Corrupted Decrepit Sewers`` or ``Corrupted Sand-Swept Tomb``"""

class Dungeon(CorkusBase):
    """Represents a `Dungeon <https://wynncraft.fandom.com/wiki/Dungeons>`_ completed by a :py:class:`Player`"""
    @property
    def name(self) -> str:
        """Name of the dungeon  like ``Decrepit Sewers``, ``Galleon's Graveyard`` or ``Fallen Factory``."""
        return self._attributes.get("name", "")

    @property
    def type(self) -> DungeonType:
        """Type of the dungeon."""
        if self.name.startswith("Corrupted"):
            return DungeonType.CORRUPTED
        elif self.name in (
            "Zombie",
            "Animal",
            "Skeleton",
            "Spider",
            "Silverfish",):
            return DungeonType.REMOVED
        elif self.name in (
            "Jungle",
            "Ice",
            "Ocean"):
            return DungeonType.REMOVED_MINI
        elif self.name in (
            "Decrepit Sewers",
            "Infested Pit",
            "Ice Barrows",
            "Lost Sanctuary",
            "Sand-Swept Tomb",
            "Underworld Crypt",
            "Undergrowth Ruins",
            "Eldritch Outlook",
            "Galleon's Graveyard",
            "Fallen Factory"):
            return DungeonType.STANDARD
        else:
            raise ValueError(f"Invalid dungeon: {self.name}")

    @property
    def completed(self) -> int:
        """Total runs completed by the player. Failed runs are not counted."""
        return self._attributes.get("completed", 0)

    def __repr__(self) -> str:
        return f"<Dungeon name={self.name!r} completed={self.completed}>"



from __future__ import annotations
from typing import Optional

from .endpoint import Endpoint
from corkus.utils.request import APIVersion
from corkus.objects import OnlinePlayers

class NetworkEndpoint(Endpoint):
    async def players_sum(self, timeout: Optional[int] = None) -> int:
        """Get number of online players across all servers.

        :param timeout: Optionally override default timeout."""
        response = await self._request.get(
            version = APIVersion.V1,
            parameters = "onlinePlayersSum",
            timeout = timeout
        )
        return response.get("players_online", 0)

    async def online_players(self, timeout: Optional[int] = None) -> OnlinePlayers:
        """List all running servers and players that are online on them.

        :param timeout: Optionally override default timeout.
        """
        response = await self._request.get(
            version = APIVersion.V1,
            parameters = "onlinePlayers",
            timeout = timeout
        )

        del response["request"]

        # api sometimes returns such server, i guess just bug
        # "": []
        if "" in response:
            del response[""]

        return OnlinePlayers(self._corkus, response)



from __future__ import annotations
from typing import TYPE_CHECKING, Optional
from .partial_base import PartialBase

if TYPE_CHECKING:
    from corkus import Corkus

class LevelRange(PartialBase):
    """Range of levels where crafted item is most effective."""

    def __init__(self, corkus: Corkus, *,
        attributes: Optional[dict] = None,
        min: Optional[int] = None,
        max: Optional[int] = None
        ):
        super().__init__(corkus)

        if attributes is None:
            attributes = {}

        self._min = attributes.get("minimum") or min or 0
        self._max = attributes.get("maximum") or max or 0

    @property
    def min(self) -> int:
        """Minimal combat level needed in order for this item to be used."""
        return self._min

    @property
    def max(self) -> int:
        """Recommended maximal combat level for this item."""
        return self._max

    def __repr__(self) -> str:
        return f"<LevelRange min={self.min} max={self.max}>"



from __future__ import annotations
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from corkus import Corkus

class PartialBase:
    def __init__(self, corkus: Corkus):
        self._corkus = corkus



from __future__ import annotations
from enum import Enum

class Timeframe(Enum):
    """List of timeframes available for some of leaderboard endpoints."""

    ALL_TIME = "alltime"
    WEEKLY = "weekly"

class LogicSymbol(Enum):
    """LogicSymbols are use in more advanced search calls in the API.

    .. py:currentmodule:: corkus.objects

    If you want to search for :py:class:`Ingredient` that require
    :py:attr:`ProfessionType.ARMOURING` AND :py:attr:`ProfessionType.SCRIBING` at the same time,
    you use :py:attr:`AND`

    If you want to search for :py:class:`Ingredient` that require
    :py:attr:`ProfessionType.ARMOURING` OR :py:attr:`ProfessionType.SCRIBING`
    you use :py:attr:`OR`. It will result all of the results that
    have any of provided skills."""

    AND = "&"
    """Return list of results that match **ALL OF THE PROVIDED ARGUMENTS**."""

    OR = "^"
    """Return list of results that match **ONE OR MORE OF THE PROVIDED ARGUMENTS**."""


class ProfessionType(Enum):
    """List of Gathering and Crafting `Professions <https://wynncraft.fandom.com/wiki/Professions>`_ plus :py:attr:`COMBAT`."""

    COMBAT = "COMBAT"
    ALCHEMISM = "ALCHEMISM"
    ARMOURING = "ARMOURING"
    COOKING = "COOKING"
    JEWELING = "JEWELING"
    SCRIBING = "SCRIBING"
    TAILORING = "TAILORING"
    WEAPONSMITHING = "WEAPONSMITHING"
    WOODWORKING = "WOODWORKING"
    MINING = "MINING"
    WOODCUTTING = "WOODCUTTING"
    FARMING = "FARMING"
    FISHING = "FISHING"

class ItemType(Enum):
    """Crafted and regular items."""

    HELMET = "HELMET"
    CHESTPLATE = "CHESTPLATE"
    LEGGINGS = "LEGGINGS"
    BOOTS = "BOOTS"
    RING = "RING"
    NECKLACE = "NECKLACE"
    BRACELET = "BRACELET"
    DAGGER = "DAGGER"
    SPEAR = "SPEAR"
    BOW = "BOW"
    WAND = "WAND"
    RELIK = "RELIK"

    SCROLL = "SCROLL"
    """Crafted item."""
    POTION = "POTION"
    """Crafted item."""
    FOOD = "FOOD"
    """Crafted item."""

class ItemCategory(Enum):
    """Crafted and regular items."""

    ARMOUR = "ARMOUR"
    WEAPON = "WEAPON"
    ACCESSORY = "ACCESSORY"
    COMSUMABLE = "COMSUMABLE"

    @staticmethod
    def from_type(type: ItemType) -> "ItemCategory":
        """Convert :py:class:`ItemType` intro :py:class:`ItemCategory`."""
        if type in (ItemType.HELMET, ItemType.CHESTPLATE, ItemType.LEGGINGS, ItemType.BOOTS):
            return ItemCategory.ARMOUR
        elif type in (ItemType.RING, ItemType.NECKLACE, ItemType.BRACELET):
            return ItemCategory.ACCESSORY
        elif type in (ItemType.DAGGER, ItemType.SPEAR, ItemType.BOW, ItemType.WAND, ItemType.RELIK):
            return ItemCategory.WEAPON
        elif type in (ItemType.SCROLL, ItemType.POTION, ItemType.FOOD):
            return ItemCategory.COMSUMABLE
        else:
            raise ValueError(f"invalid type: {type}")



from __future__ import annotations
from typing import List, Union
import time
import logging

logger = logging.getLogger("corkus.cache")

class CacheElement():
    def __init__(
        self,
        endpoint_url: str,
        valid_timestamp: int,
        content: dict
    ) -> None:
        self._url = endpoint_url
        self._valid_timestamp = valid_timestamp
        self._content = content

    @property
    def url(self) -> str:
        return self._url

    @property
    def valid_timestamp(self) -> int:
        return self._valid_timestamp

    @property
    def content(self) -> dict:
        return self._content

    def __repr__(self) -> str:
        return f"<CacheElement url={self.url!r} valid_timestamp={self.valid_timestamp}>"

class CorkusCache:
    def __init__(self) -> None:
        self._cache: List[CacheElement] = []

    @property
    def content(self) -> List[CacheElement]:
        for item in self._cache:
            if item.valid_timestamp <= time.time():
                self._cache.remove(item)
        return self._cache

    def get(self, url: str) -> Union[CacheElement, None]:
        element = next((i for i in self.content if i.url == url), None)
        if element is None:
            logger.debug(f"Not found in cache: {url}")
        else:
            logger.debug(f"Serving from cache: {url}")
        return element

    def add(self, url: str, headers: dict, content: dict) -> None:
        cache_header = headers.get("cache-control", None)
        if cache_header is None:
            logger.debug(f"No cache-control for: {url} - default to 600")
            cache_header = "max-age=600"
        seconds = cache_header.replace("max-age=", "")
        if not seconds.isdigit():
            logger.debug(f"Invalid cache-control ({seconds}) for: {url} - default to 600")
            seconds = 600
        logger.debug(f"Caching: {url} - for {seconds} seconds")

        self._cache.append(CacheElement(url, int(time.time()) + int(seconds), content))



from __future__ import annotations
from datetime import datetime
from typing import Union

from .base import CorkusBase
from .partial_guild import PartialGuild
from .territory_location import TerritoryLocation

class Territory(CorkusBase):
    """Territories are areas which may be claimed by a :py:class:`Guild` to receive benefits."""

    @property
    def name(self) -> str:
        """The name of the territory."""
        return self._attributes.get("territory", None)

    @property
    def guild(self) -> Union[PartialGuild, None]:
        """Guild that currently holds the territory."""
        guid_name = self._attributes.get("guild", None)
        if guid_name is None or guid_name == "Nobody":
            return None
        else:
            return PartialGuild(self._corkus, guid_name)

    @property
    def acquired(self) -> datetime:
        """Datetime when the territory was acquired."""
        return datetime.strptime(self._attributes.get("acquired", "1970-01-01 00:00:00"), "%Y-%m-%d %H:%M:%S")

    @property
    def attacker(self) -> Union[PartialGuild, None]:
        """Guild that is currently attacking the territory."""
        guid_name = self._attributes.get("attacker", None)
        if guid_name is None:
            return None
        else:
            return PartialGuild(self._corkus, guid_name)

    @property
    def is_attacked(self) -> bool:
        """Is this territory currently under attack."""
        return self.attacker is not None

    @property
    def location(self) -> TerritoryLocation:
        """The location of this territory."""
        return TerritoryLocation(self._corkus, self._attributes.get("location", {}))

    def __repr__(self) -> str:
        return f"<Territory name={self.name!r} guild={self.guild}>"



from setuptools import setup
import re

requirements = []
with open('requirements.txt') as f:
    requirements = f.read().splitlines()

version = ''
with open('corkus/version.py') as f:
    version = re.search(r'^__version__\s*=\s*[\'"]([^\'"]*)[\'"]', f.read(), re.MULTILINE).group(1)

if not version:
    raise RuntimeError('version is not set')

readme = ''
with open('README.md', encoding="utf8") as f:
    readme = f.read()

setup(
    name='corkus.py',
    author='MrBartusek',
    url='https://github.com/MrBartusek/corkus.py',
    project_urls={
        "Documentation": "https://corkuspy.readthedocs.io/en/latest/",
        "Issue tracker": "https://github.com/MrBartusek/corkus.py/issues"
    },
    package_data={"corkus": ["py.typed"]},
    packages=[
        "corkus",
        "corkus.objects",
        "corkus.endpoints",
        "corkus.utils",
        "corkus.data"
    ],
    version=version,
    license='MIT',
    description='A Python wrapper for the Wynncraft API',
    long_description=readme,
    long_description_content_type="text/markdown",
    include_package_data=True,
    install_requires=requirements,
    python_requires='>=3.8.0',
    classifiers=[
    'Development Status :: 4 - Beta',
    'License :: OSI Approved :: MIT License',
    'Intended Audience :: Developers',
    'Natural Language :: English',
    'Operating System :: OS Independent',
    'Programming Language :: Python :: 3',
    'Programming Language :: Python :: 3.8',
    'Programming Language :: Python :: 3.9',
    'Topic :: Internet',
    'Topic :: Games/Entertainment',
    'Topic :: Software Development :: Libraries',
    'Topic :: Software Development :: Libraries :: Python Modules',
    'Topic :: Utilities',
    ]
)



from corkus.objects import LogicSymbol

class Utils:
    @staticmethod
    def build_complex_query(symbol: LogicSymbol, **kwargs: str):
        props = []
        for key, value in kwargs.items():
            if value is None: continue
            props.append(f"{key.lower()}<{value}>")
        if len(props) == 0:
            raise ValueError("at least one keyword argument is reqired for complex search routes")
        return symbol.value + ",".join(props)
